@@ -55,7 +55,23 @@ inline int str_startswith(const char *head, const char *worm)
     return !strncmp(head, worm, strlen(head));
 }
 
-static bool analyze_program(const char *name, CompileJob &job)
+/* Some files should always be built locally... */
+static bool
+should_always_build_locally(const string &filename)
+{
+    string p;
+
+    p = find_basename(filename);
+
+    if (str_startswith("conftest.", p.c_str())
+        || str_startswith("tmp.conftest.", p.c_str())) {
+        return true;
+    }
+
+    return false;
+}
+
+static bool analyze_program(const char *name, CompileJob &job, bool& icerun)
 {
     string compiler_name = find_basename(name);
 
@@ -79,7 +95,11 @@ static bool analyze_program(const char *name, CompileJob &job)
     bool vclang = compiler_name.find("clang-") != string::npos;
     bool vclangpp = compiler_name.find("clang++-") != string::npos;
 
-    if ((suffix == "++") || (suffix == "CC") || vgpp || vclangpp) {
+    if( icerun ) {
+        job.setLanguage(CompileJob::Lang_Custom);
+        log_info() << "icerun, running locally." << endl;
+        return true;
+    } else if ((suffix == "++") || (suffix == "CC") || vgpp || vclangpp) {
         job.setLanguage(CompileJob::Lang_CXX);
     } else if ((suffix == "cc") || vgcc || vclang) {
         job.setLanguage(CompileJob::Lang_C);
@@ -88,6 +108,7 @@ static bool analyze_program(const char *name, CompileJob &job)
     } else {
         job.setLanguage(CompileJob::Lang_Custom);
         log_info() << "custom command, running locally." << endl;
+        icerun = true;
         return true;
     }
 
@@ -106,67 +127,105 @@ static bool is_argument_with_space(const char* argument)
     //         -segcreate <arg1> <arg2> <arg3>
     //         -segprot <arg1> <arg2> <arg3>
     //       Move some arguments to Arg_Cpp or Arg_Local
-    if (str_equal("-dyld-prefix", argument)
-        || str_equal("-gcc-toolchain", argument)
-        || str_equal("--param", argument)
-        || str_equal("--sysroot", argument)
-        || str_equal("--system-header-prefix", argument)
-        || str_equal("-target", argument)
-        || str_equal("--assert", argument)
-        || str_equal("--allowable_client", argument)
-        || str_equal("-arch", argument)
-        || str_equal("-arch_only", argument)
-        || str_equal("-arcmt-migrate-report-output", argument)
-        || str_equal("--prefix", argument)
-        || str_equal("-bundle_loader", argument)
-        || str_equal("-dependency-dot", argument)
-        || str_equal("-dependency-file", argument)
-        || str_equal("-dylib_file", argument)
-        || str_equal("-exported_symbols_list", argument)
-        || str_equal("--bootclasspath", argument)
-        || str_equal("--CLASSPATH", argument)
-        || str_equal("--classpath", argument)
-        || str_equal("--resource", argument)
-        || str_equal("--encoding", argument)
-        || str_equal("--extdirs", argument)
-        || str_equal("-filelist", argument)
-        || str_equal("-fmodule-implementation-of", argument)
-        || str_equal("-fmodule-name", argument)
-        || str_equal("-fmodules-user-build-path", argument)
-        || str_equal("-fnew-alignment", argument)
-        || str_equal("-force_load", argument)
-        || str_equal("--output-class-directory", argument)
-        || str_equal("-framework", argument)
-        || str_equal("-frewrite-map-file", argument)
-        || str_equal("-ftrapv-handler", argument)
-        || str_equal("-image_base", argument)
-        || str_equal("-init", argument)
-        || str_equal("-install_name", argument)
-        || str_equal("-lazy_framework", argument)
-        || str_equal("-lazy_library", argument)
-        || str_equal("-meabi", argument)
-        || str_equal("-mhwdiv", argument)
-        || str_equal("-mllvm", argument)
-        || str_equal("-module-dependency-dir", argument)
-        || str_equal("-mthread-model", argument)
-        || str_equal("-multiply_defined", argument)
-        || str_equal("-multiply_defined_unused", argument)
-        || str_equal("-rpath", argument)
-        || str_equal("--rtlib", argument)
-        || str_equal("-seg_addr_table", argument)
-        || str_equal("-seg_addr_table_filename", argument)
-        || str_equal("-segs_read_only_addr", argument)
-        || str_equal("-segs_read_write_addr", argument)
-        || str_equal("-serialize-diagnostics", argument)
-        || str_equal("-std", argument)
-        || str_equal("--stdlib", argument)
-        || str_equal("--force-link", argument)
-        || str_equal("-umbrella", argument)
-        || str_equal("-unexported_symbols_list", argument)
-        || str_equal("-weak_library", argument)
-        || str_equal("-weak_reference_mismatches", argument)) {
+    static const char* const arguments[] = {
+        "-dyld-prefix",
+        "-gcc-toolchain",
+        "--param",
+        "--sysroot",
+        "--system-header-prefix",
+        "-target",
+        "--assert",
+        "--allowable_client",
+        "-arch",
+        "-arch_only",
+        "-arcmt-migrate-report-output",
+        "--prefix",
+        "-bundle_loader",
+        "-dependency-dot",
+        "-dependency-file",
+        "-dylib_file",
+        "-exported_symbols_list",
+        "--bootclasspath",
+        "--CLASSPATH",
+        "--classpath",
+        "--resource",
+        "--encoding",
+        "--extdirs",
+        "-filelist",
+        "-fmodule-implementation-of",
+        "-fmodule-name",
+        "-fmodules-user-build-path",
+        "-fnew-alignment",
+        "-force_load",
+        "--output-class-directory",
+        "-framework",
+        "-frewrite-map-file",
+        "-ftrapv-handler",
+        "-image_base",
+        "-init",
+        "-install_name",
+        "-lazy_framework",
+        "-lazy_library",
+        "-meabi",
+        "-mhwdiv",
+        "-mllvm",
+        "-module-dependency-dir",
+        "-mthread-model",
+        "-multiply_defined",
+        "-multiply_defined_unused",
+        "-rpath",
+        "--rtlib",
+        "-seg_addr_table",
+        "-seg_addr_table_filename",
+        "-segs_read_only_addr",
+        "-segs_read_write_addr",
+        "-serialize-diagnostics",
+        "-std",
+        "--stdlib",
+        "--force-link",
+        "-umbrella",
+        "-unexported_symbols_list",
+        "-weak_library",
+        "-weak_reference_mismatches",
+        "-B",
+        "-D",
+        "-U",
+        "-I",
+        "-i",
+        "--include-directory",
+        "-L",
+        "-l",
+        "--library-directory",
+        "-MF",
+        "-MT",
+        "-MQ",
+        "-cxx-isystem",
+        "-c-isystem",
+        "-idirafter",
+        "--include-directory-after",
+        "-iframework",
+        "-iframeworkwithsysroot",
+        "-imacros",
+        "-imultilib",
+        "-iprefix",
+        "--include-prefix",
+        "-iquote",
+        "-isysroot",
+        "-isystem",
+        "-isystem-after",
+        "-ivfsoverlay",
+        "-iwithprefix",
+        "--include-with-prefix",
+        "--include-with-prefix-after",
+        "-iwithprefixbefore",
+        "--include-with-prefix-before",
+        "-iwithsysroot"
+    };
 
-        return true;
+    for( size_t i = 0; i < sizeof( arguments ) / sizeof( arguments[ 0 ] ); ++i ) {
+        if (str_equal( arguments[ i ], argument)) {
+            return true;
+        }
     }
 
     return false;
@@ -189,24 +248,22 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
 #endif
 
     bool had_cc = (job.compilerName().size() > 0);
-    bool always_local = analyze_program(had_cc ? job.compilerName().c_str() : argv[0], job);
+    bool always_local = analyze_program(had_cc ? job.compilerName().c_str() : argv[0], job, icerun);
     bool seen_c = false;
     bool seen_s = false;
     bool seen_mf = false;
     bool seen_md = false;
     bool seen_split_dwarf = false;
+    bool seen_target = false;
+    bool wunused_macros = false;
+    bool seen_arch = false;
+    bool seen_pedantic = false;
     // if rewriting includes and precompiling on remote machine, then cpp args are not local
     Argument_Type Arg_Cpp = compiler_only_rewrite_includes(job) ? Arg_Rest : Arg_Local;
 
     explicit_color_diagnostics = false;
     explicit_no_show_caret = false;
 
-    if (icerun) {
-        always_local = true;
-        job.setLanguage(CompileJob::Lang_Custom);
-        log_info() << "icerun, running locally." << endl;
-    }
-
     for (int i = had_cc ? 2 : 1; argv[i]; i++) {
         const char *a = argv[i];
 
@@ -217,11 +274,15 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                 always_local = true;
                 args.append(a, Arg_Local);
                 log_info() << "preprocessing, building locally" << endl;
-            } else if (!strncmp(a, "-fdump", 6) || !strcmp(a, "-combine")) {
+            } else if (!strncmp(a, "-fdump", 6)
+                       || !strcmp(a, "-combine")
+                       || !strcmp(a, "-fsyntax-only")) {
                 always_local = true;
                 args.append(a, Arg_Local);
                 log_info() << "argument " << a << ", building locally" << endl;
-            } else if (!strcmp(a, "-MD") || !strcmp(a, "-MMD")) {
+            } else if (!strcmp(a, "-MD") || !strcmp(a, "-MMD") ||
+                       str_startswith("-Wp,-MD", a) ||
+                       str_startswith("-Wp,-MMD", a)) {
                 seen_md = true;
                 args.append(a, Arg_Local);
                 /* These two generate dependencies as a side effect.  They
@@ -252,6 +313,7 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
             } else if (str_equal("--param", a)) {
                 args.append(a, Arg_Remote);
 
+                assert( is_argument_with_space( a ));
                 /* skip next word, being option argument */
                 if (argv[i + 1]) {
                     args.append(argv[++i], Arg_Remote);
@@ -265,6 +327,7 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                 log_info() << "argument " << a << ", building locally" << endl;
 
                 if (str_equal(a, "-B")) {
+                    assert( is_argument_with_space( a ));
                     /* skip next word, being option argument */
                     if (argv[i + 1]) {
                         args.append(argv[++i], Arg_Local);
@@ -348,7 +411,18 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                     ++i;
                     args.append(opt, Arg_Rest);
                     if (str_equal(opt, "c++") || str_equal(opt, "c")) {
-                        CompileJob::Language lang = str_equal(opt, "c++") ? CompileJob::Lang_CXX : CompileJob::Lang_C;
+                        CompileJob::Language lang = CompileJob::Lang_Custom;
+                        if( str_equal(opt, "c")) {
+                            lang = CompileJob::Lang_C;
+                        } else if( str_equal(opt, "c++")) {
+                            lang = CompileJob::Lang_CXX;
+                        } else if( str_equal(opt, "objective-c")) {
+                            lang = CompileJob::Lang_OBJC;
+                        } else if( str_equal(opt, "objective-c++")) {
+                            lang = CompileJob::Lang_OBJCXX;
+                        } else {
+                            continue;
+                        }
                         job.setLanguage(lang); // will cause -x used remotely twice, but shouldn't be a problem
                         unsupported = false;
                     }
@@ -396,6 +470,7 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                 /* This has a duplicate meaning. it can either include a file
                    for preprocessing or a precompiled header. decide which one.  */
 
+                args.append(a, Arg_Local);
                 if (argv[i + 1]) {
                     ++i;
                     std::string p = argv[i];
@@ -404,9 +479,8 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                     if (dot_index != string::npos) {
                         string ext = p.substr(dot_index + 1);
 
-                        if (ext[0] != 'h' && ext[0] != 'H'
-                            && (access(p.c_str(), R_OK)
-                                || access((p + ".gch").c_str(), R_OK))) {
+                        if (ext[0] != 'h' && ext[0] != 'H' && access(p.c_str(), R_OK) < 0
+                            && access((p + ".gch").c_str(), R_OK) < 0) {
                             log_info() << "include file or gch file for argument " << a << " " << p
                                        << " missing, building locally" << endl;
                             always_local = true;
@@ -416,20 +490,24 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                         always_local = true;    /* Included file is not header.suffix or header.suffix.gch! */
                     }
 
-                    args.append(a, Arg_Local);
                     args.append(argv[i], Arg_Local);
                 }
             } else if (str_equal("-include-pch", a)) {
                 /* Clang's precompiled header, it's probably not worth it sending the PCH file. */
+                args.append(a, Arg_Local);
                 if (argv[i + 1]) {
                     ++i;
+                    args.append(argv[i], Arg_Local);
+                    if(access(argv[i], R_OK) < 0) {
+                        log_info() << "pch file for argument " << a << " " << argv[i]
+                                   << " missing, building locally" << endl;
+                        always_local = true;
+                    }
                 }
-
-                always_local = true;
-                log_info() << "argument " << a << ", building locally" << endl;
             } else if (str_equal("-D", a) || str_equal("-U", a)) {
                 args.append(a, Arg_Cpp);
 
+                assert( is_argument_with_space( a ));
                 /* skip next word, being option argument */
                 if (argv[i + 1]) {
                     ++i;
@@ -467,6 +545,7 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                        || str_equal("-iwithsysroot", a)) {
                 args.append(a, Arg_Local);
 
+                assert( is_argument_with_space( a ));
                 /* skip next word, being option argument */
                 if (argv[i + 1]) {
                     ++i;
@@ -495,6 +574,8 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                        || str_equal("-MG", a)
                        || str_equal("-MP", a)) {
                 args.append(a, Arg_Local);
+            } else if (str_equal("-Wmissing-include-dirs", a)) {
+                args.append(a, Arg_Local);
             } else if (str_equal("-fno-color-diagnostics", a)) {
                 explicit_color_diagnostics = true;
                 args.append(a, Arg_Rest);
@@ -516,21 +597,32 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
             } else if (str_equal("-fno-diagnostics-show-caret", a)) {
                 explicit_no_show_caret = true;
                 args.append(a, Arg_Rest);
-            } else if (str_equal("-flto", a)) {
-                // pointless when preprocessing, and Clang would emit a warning
-                args.append(a, Arg_Remote);
-            } else if (str_startswith("-fplugin=", a)) {
-                string file = a + strlen("-fplugin=");
+            } else if (str_equal("-fdiagnostics-show-caret", a)) {
+                explicit_no_show_caret = false;
+                args.append(a, Arg_Rest);
+            } else if (str_startswith("-fplugin=", a)
+                       || str_startswith("-fsanitize-blacklist=", a)
+                       || str_startswith("-fprofile-sample-use=", a)) {
+                const char* prefix = NULL;
+                static const char* const prefixes[] = { "-fplugin=", "-fsanitize-blacklist=", "-fprofile-sample-use=" };
+                for( size_t pref = 0; pref < sizeof(prefixes)/sizeof(prefixes[0]); ++pref) {
+                    if( str_startswith(prefixes[pref], a)) {
+                        prefix = prefixes[pref];
+                        break;
+                    }
+                }
+                assert( prefix != NULL );
+                string file = a + strlen(prefix);
 
                 if (access(file.c_str(), R_OK) == 0) {
                     file = get_absfilename(file);
                     extrafiles->push_back(file);
                 } else {
                     always_local = true;
-                    log_info() << "plugin for argument " << a << " missing, building locally" << endl;
+                    log_info() << "file for argument " << a << " missing, building locally" << endl;
                 }
 
-                args.append("-fplugin=" + file, Arg_Rest);
+                args.append(prefix + file, Arg_Rest);
             } else if (str_equal("-Xclang", a)) {
                 if (argv[i + 1]) {
                     ++i;
@@ -561,6 +653,37 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                         args.append(p, Arg_Rest);
                     }
                 }
+            } else if (str_equal("-target", a)) {
+                seen_target = true;
+                args.append(a, Arg_Rest);
+                if (argv[i + 1]) {
+                    args.append(argv[++i], Arg_Rest);
+                }
+            } else if (str_startswith("--target=", a)) {
+                seen_target = true;
+                args.append(a, Arg_Rest);
+            } else if (str_equal("-Wunused-macros", a)) {
+                wunused_macros = true;
+                args.append(a, Arg_Rest);
+            } else if (str_equal("-Wno-unused-macros", a)) {
+                wunused_macros = false;
+                args.append(a, Arg_Rest);
+            } else if (str_equal("-pedantic", a)) {
+                seen_pedantic = true;
+                args.append(a, Arg_Rest);
+            } else if (str_equal("-pedantic-errors", a)) {
+                seen_pedantic = true;
+                args.append(a, Arg_Rest);
+            } else if (str_equal("-arch", a)) {
+                if( seen_arch ) {
+                    log_info() << "multiple -arch options, building locally" << endl;
+                    always_local = true;
+                }
+                seen_arch = false;
+                args.append(a, Arg_Rest);
+                if (argv[i + 1]) {
+                    args.append(argv[++i], Arg_Rest);
+                }
             } else {
                 args.append(a, Arg_Rest);
 
@@ -589,6 +712,7 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
 
         args.append("-S", Arg_Remote);
     } else {
+        assert( seen_c );
         args.append("-c", Arg_Remote);
         if (seen_split_dwarf) {
             job.setDwarfFissionEnabled(true);
@@ -625,6 +749,11 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
                 job.setInputFile(it->first);
                 ifile = it->first;
                 it = args.erase(it);
+                if (should_always_build_locally(ifile)) {
+                    log_info() << "autoconf tests are run locally: "
+                               << ifile << endl;
+                    always_local = true;
+                }
             } else {
                 log_info() << "found another non option on command line. Two input files? "
                            << it->first << endl;
@@ -651,10 +780,11 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
 
 #endif
                 job.setLanguage(CompileJob::Lang_CXX);
-            } else if (ext == "mi" || ext == "m"
-                       || ext == "mii" || ext == "mm"
-                       || ext == "M") {
+            } else if (ext == "mi" || ext == "m") {
                 job.setLanguage(CompileJob::Lang_OBJC);
+            } else if (ext == "mii" || ext == "mm"
+                       || ext == "M") {
+                job.setLanguage(CompileJob::Lang_OBJCXX);
             } else if (ext == "s" || ext == "S" // assembler
                        || ext == "ads" || ext == "adb" // ada
                        || ext == "f" || ext == "for" // fortran
@@ -696,28 +826,62 @@ bool analyse_argv(const char * const *argv, CompileJob &job, bool icerun, list<s
             }
         }
 
-    } else {
+    } else { // always_local
         job.setInputFile(string());
     }
 
     struct stat st;
 
-    if (ofile.empty() || (!stat(ofile.c_str(), &st) && !S_ISREG(st.st_mode))) {
-        if (!always_local) {
+    if( !always_local ) {
+        if (ofile.empty() || (!stat(ofile.c_str(), &st) && !S_ISREG(st.st_mode))) {
             log_info() << "output file empty or not a regular file, building locally" << endl;
+            always_local = true;
         }
-        always_local = true;
     }
 
     // redirecting compiler's output will turn off its automatic coloring, so force it
     // when it would be used, unless explicitly set
-    if (compiler_has_color_output(job) && !explicit_color_diagnostics) {
+    if (!icerun && compiler_has_color_output(job) && !explicit_color_diagnostics) {
         if (compiler_is_clang(job))
             args.append("-fcolor-diagnostics", Arg_Rest);
         else
             args.append("-fdiagnostics-color", Arg_Rest); // GCC
     }
 
+    // -Wunused-macros is tricky with remote preprocessing. GCC's -fdirectives-only outright
+    // refuses to work if -Wunused-macros is given, and Clang's -frewrite-includes is buggy
+    // if -Wunused-macros is used (https://bugs.llvm.org/show_bug.cgi?id=15614). It's a question
+    // if this could possibly even work, given that macros may be used as filenames for #include directives.
+    if( wunused_macros ) {
+        job.setBlockRewriteIncludes(true);
+    }
+
+    // -pedantic doeesn't work with remote preprocessing
+    if( seen_pedantic ) {
+        job.setBlockRewriteIncludes(true);
+    }
+
+    if( !always_local && compiler_only_rewrite_includes(job) && !compiler_is_clang(job)) {
+        // Inject this, so that remote compilation uses -fpreprocessed -fdirectives-only
+        args.append("-fdirectives-only", Arg_Remote);
+    }
+
+    if( !always_local && !seen_target && compiler_is_clang(job)) {
+        // With gcc each binary can compile only for one target, so cross-compiling is just
+        // a matter of using the proper cross-compiler remotely and it will automatically
+        // compile for the given platform. However one clang binary can compile for many
+        // platforms and so when cross-compiling it would by default compile for the remote
+        // host platform. Therefore explicitly ask for our platform.
+        string default_target = clang_get_default_target(job);
+        if( !default_target.empty()) {
+            args.append("-target", Arg_Remote);
+            args.append(default_target, Arg_Remote);
+        } else {
+            always_local = true;
+            log_error() << "failed to read default clang host platform, building locally" << endl;
+        }
+    }
+
     job.setFlags(args);
     job.setOutputFile(ofile);
 
