@@ -63,7 +63,7 @@ namespace
 
 struct CharBufferDeleter {
     char *buf;
-    CharBufferDeleter(char *b) : buf(b) {}
+    explicit CharBufferDeleter(char *b) : buf(b) {}
     ~CharBufferDeleter() {
         free(buf);
     }
@@ -125,7 +125,7 @@ parse_icecc_version(const string &target_platform, const string &prefix)
             continue;
         }
 
-        if (::access(version.c_str(), R_OK)) {
+        if (::access(version.c_str(), R_OK) < 0) {
             log_error() << "$ICECC_VERSION has to point to an existing file to be installed " << version << endl;
             continue;
         }
@@ -345,7 +345,7 @@ static void receive_file(const string& output_file, MsgChannel* cserver)
         if (msg->type != M_FILE_CHUNK) {
             unlink(tmp_file.c_str());
             delete msg;
-            throw client_error(20, "Error 20 - unexpcted message");
+            throw client_error(20, "Error 20 - unexpected message");
         }
 
         FileChunkMsg *fcmsg = dynamic_cast<FileChunkMsg*>(msg);
@@ -481,6 +481,13 @@ static int build_remote_int(CompileJob &job, UseCSMsg *usecs, MsgChannel *local_
             throw client_error(26, "Error 26 - environment on " + hostname + " cannot be verified");
         }
 
+        // Older remotes don't set properly -x argument.
+        if(( job.language() == CompileJob::Lang_OBJC || job.language() == CompileJob::Lang_OBJCXX )
+            && !IS_PROTOCOL_38(cserver)) {
+            job.appendFlag( "-x", Arg_Remote );
+            job.appendFlag( job.language() == CompileJob::Lang_OBJC ? "objective-c" : "objective-c++", Arg_Remote );
+        }
+
         CompileFileMsg compile_file(&job);
         {
             log_block b("send compile_file");
@@ -523,6 +530,7 @@ static int build_remote_int(CompileJob &job, UseCSMsg *usecs, MsgChannel *local_
             if (shell_exit_status(status) != 0) {   // failure
                 delete cserver;
                 cserver = 0;
+                log_warning() << "call_cpp process failed with exit status " << shell_exit_status(status) << endl;
                 return shell_exit_status(status);
             }
         } else {
@@ -574,6 +582,7 @@ static int build_remote_int(CompileJob &job, UseCSMsg *usecs, MsgChannel *local_
             if ((!crmsg->out.empty() || !crmsg->err.empty()) && output_needs_workaround(job)) {
                 delete crmsg;
                 log_info() << "command needs stdout/stderr workaround, recompiling locally" << endl;
+                log_info() << "(set ICECC_CARET_WORKAROUND=0 to override)" << endl;
                 throw remote_error(102, "Error 102 - command needs stdout/stderr workaround, recompiling locally");
             }
 
@@ -606,7 +615,7 @@ static int build_remote_int(CompileJob &job, UseCSMsg *usecs, MsgChannel *local_
     } catch (...) {
         // Handle pending status messages, if any.
         if(cserver) {
-            while(Msg* msg = cserver->get_msg(0)) {
+            while(Msg* msg = cserver->get_msg(0, true)) {
                 if(msg->type == M_STATUS_TEXT)
                     log_error() << "Remote status (compiled on " << cserver->name << "): "
                                 << static_cast<StatusTextMsg*>(msg)->text << endl;
@@ -748,19 +757,19 @@ int build_remote(CompileJob &job, MsgChannel *local_daemon, const Environments &
     int torepeat = 1;
     bool has_split_dwarf = job.dwarfFissionEnabled();
 
-    // older compilers do not support the options we need to make it reproducible
-#if defined(__GNUC__) && ( ( (__GNUC__ == 3) && (__GNUC_MINOR__ >= 3) ) || (__GNUC__ >=4) )
-
     if (!compiler_is_clang(job)) {
         if (rand() % 1000 < permill) {
             torepeat = 3;
         }
     }
 
-#endif
-
-    trace() << job.inputFile() << " compiled " << torepeat << " times on "
-            << job.targetPlatform() << "\n";
+    if( torepeat == 1 ) {
+        trace() << "preparing " << job.inputFile() << " to be compiled for "
+                << job.targetPlatform() << "\n";
+    } else {
+        trace() << "preparing " << job.inputFile() << " to be compiled " << torepeat << " times for "
+                << job.targetPlatform() << "\n";
+    }
 
     map<string, string> versionfile_map, version_map;
     Environments envs = rip_out_paths(_envs, version_map, versionfile_map);
@@ -801,11 +810,16 @@ int build_remote(CompileJob &job, MsgChannel *local_daemon, const Environments &
         UseCSMsg *usecs = get_server(local_daemon);
         int ret;
 
-        if (!maybe_build_local(local_daemon, usecs, job, ret))
-            ret = build_remote_int(job, usecs, local_daemon,
-                                   version_map[usecs->host_platform],
-                                   versionfile_map[usecs->host_platform],
-                                   0, true);
+        try {
+            if (!maybe_build_local(local_daemon, usecs, job, ret))
+                ret = build_remote_int(job, usecs, local_daemon,
+                                       version_map[usecs->host_platform],
+                                       versionfile_map[usecs->host_platform],
+                                       0, true);
+        } catch(...) {
+            delete usecs;
+            throw;
+        }
 
         delete usecs;
         return ret;
@@ -827,6 +841,7 @@ int build_remote(CompileJob &job, MsgChannel *local_daemon, const Environments &
         waitpid(cpp_pid, &status, 0);
 
         if (shell_exit_status(status)) {   // failure
+            log_warning() << "call_cpp process failed with exit status " << shell_exit_status(status) << endl;
             ::unlink(preproc);
             return shell_exit_status(status);
         }
