@@ -32,9 +32,7 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
-#ifdef HAVE_SIGNAL_H
 #include <signal.h>
-#endif
 
 #include "comm.h"
 #include "exitcode.h"
@@ -42,50 +40,6 @@
 
 using namespace std;
 
-#if 0
-static string read_fromFILE(FILE *f)
-{
-    string output;
-
-    if (!f) {
-        log_error() << "no pipe " << strerror(errno) << endl;
-        return output;
-    }
-
-    char buffer[100];
-
-    while (!feof(f)) {
-        size_t bytes = fread(buffer, 1, 99, f);
-        buffer[bytes] = 0;
-        output += buffer;
-    }
-
-    pclose(f);
-    return output;
-}
-
-static bool extract_version(string &version)
-{
-    string::size_type pos = version.find_last_of('\n');
-
-    if (pos == string::npos) {
-        return false;
-    }
-
-    while (pos + 1 == version.size()) {
-        version.resize(version.size() - 1);
-        pos = version.find_last_of('\n');
-
-        if (pos == string::npos) {
-            return false;
-        }
-    }
-
-    version = version.substr(pos + 1);
-    return true;
-}
-#endif
-
 size_t sumup_dir(const string &dir)
 {
     size_t res = 0;
@@ -133,7 +87,7 @@ static void list_target_dirs(const string &current_target, const string &targetd
     for (struct dirent *ent = readdir(envdir); ent; ent = readdir(envdir)) {
         string dirname = ent->d_name;
 
-        if (!access(string(targetdir + "/" + dirname + "/usr/bin/as").c_str(), X_OK)) {
+        if (access(string(targetdir + "/" + dirname + "/usr/bin/as").c_str(), X_OK) == 0) {
             envs.push_back(make_pair(current_target, dirname));
         }
     }
@@ -416,6 +370,7 @@ int start_create_env(const string &basedir, uid_t user_uid, gid_t user_gid,
     int pos = 0;
     argv[pos++] = BINDIR "/icecc";
     argv[pos++] = "--build-native";
+    const int first_to_free = pos;
     argv[pos++] = strdup(compiler.c_str());
 
     for (list<string>::const_iterator it = extrafiles.begin(); it != extrafiles.end(); ++it) {
@@ -428,6 +383,9 @@ int start_create_env(const string &basedir, uid_t user_uid, gid_t user_gid,
         log_error() << BINDIR "/icecc --build-native failed" << endl;
         _exit(1);
     }
+    for( int i = first_to_free; i < pos; ++i )
+        free( (void*) argv[ i ] );
+    delete[] argv;
 
     _exit(0);
 }
@@ -783,8 +741,8 @@ bool verify_env(MsgChannel *client, const string &basedir, const string &target,
 
     string dirname = basedir + "/target=" + target + "/" + env;
 
-    if (::access(string(dirname + "/bin/true").c_str(), X_OK)) {
-        error_client(client, dirname + "/bin/true is not executable");
+    if (::access(string(dirname + "/bin/true").c_str(), X_OK) < 0) {
+        error_client(client, dirname + "/bin/true is not executable, installed environment removed?");
         log_error() << "I don't have environment " << env << "(" << target << ") to verify." << endl;
         return false;
     }
@@ -805,7 +763,7 @@ bool verify_env(MsgChannel *client, const string &basedir, const string &target,
     }
 
     // child
-    reset_debug(0);
+    reset_debug();
     chdir_to_environment(client, dirname, user_uid, user_gid);
     execl("bin/true", "bin/true", (void*)NULL);
     log_perror("execl failed");
