@@ -37,9 +37,7 @@
 #include <netdb.h>
 #include <getopt.h>
 
-#ifdef HAVE_SIGNAL_H
 #include <signal.h>
-#endif
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -69,10 +67,6 @@
 #endif
 #include <netdb.h>
 
-#ifdef HAVE_SYS_RESOURCE_H
-#  include <sys/resource.h>
-#endif
-
 #ifndef RUSAGE_SELF
 #  define RUSAGE_SELF (0)
 #endif
@@ -411,13 +405,20 @@ void usage(const char *reason = 0)
         cerr << reason << endl;
     }
 
-    cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [--no-remote] [-w] [-d|--daemonize] [-l logfile] [-s <schedulerhost[:port]>]"
+    cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [--no-remote] [-d|--daemonize] [-l logfile] [-s <schedulerhost[:port]>]"
         " [-v[v[v]]] [-u|--user-uid <user_uid>] [-b <env-basedir>] [--cache-limit <MB>] [-N <node_name>]" << endl;
     exit(1);
 }
 
 struct timeval last_stat;
+
+// Initial rlimit for a compile job, measured in megabytes.  Will vary with
+// the amount of available memory.
 int mem_limit = 100;
+
+// Minimum rlimit for a compile job, measured in megabytes.
+const int min_mem_limit = 100;
+
 unsigned int max_kids = 0;
 
 size_t cache_size_limit = 100 * 1024 * 1024;
@@ -431,6 +432,12 @@ struct NativeEnvironment {
     time_t gpp_bin_timestamp;
     time_t clang_bin_timestamp;
     int create_env_pipe; // if in progress of creating the environment
+    NativeEnvironment() {
+        gcc_bin_timestamp = 0;
+        gpp_bin_timestamp = 0;
+        clang_bin_timestamp = 0;
+        create_env_pipe = 0;
+    }
 };
 
 struct Daemon {
@@ -489,7 +496,7 @@ struct Daemon {
             user_gid = getgid();
         }
 
-        envbasedir = "/tmp/icecc-envs";
+        envbasedir = "/var/tmp/icecc-envs";
         tcp_listen_fd = -1;
         unix_listen_fd = -1;
         new_client_id = 0;
@@ -515,7 +522,7 @@ struct Daemon {
     }
 
     bool reannounce_environments() __attribute_warn_unused_result__;
-    int answer_client_requests();
+    void answer_client_requests();
     bool handle_transfer_env(Client *client, Msg *msg) __attribute_warn_unused_result__;
     bool handle_transfer_env_done(Client *client);
     bool handle_get_native_env(Client *client, GetNativeEnvMsg *msg) __attribute_warn_unused_result__;
@@ -539,7 +546,7 @@ struct Daemon {
     string dump_internals() const;
     string determine_nodename();
     void determine_system();
-    bool maybe_stats(bool force = false);
+    bool maybe_stats(bool force_check = false);
     bool send_scheduler(const Msg &msg) __attribute_warn_unused_result__;
     void close_scheduler();
     bool reconnect();
@@ -608,7 +615,8 @@ bool Daemon::setup_listen_fds()
 
     myaddr.sun_family = AF_UNIX;
 
-    mode_t old_umask = -1U;
+    bool reset_umask = false;
+    mode_t old_umask = 0;
 
     if (getenv("ICECC_TEST_SOCKET") == NULL) {
 #ifdef HAVE_LIBCAP_NG
@@ -623,10 +631,11 @@ bool Daemon::setup_listen_fds()
             if(default_socket.length() > sizeof(myaddr.sun_path) - 1) {
                 log_error() << "default socket path too long for sun_path" << endl;	
             }
-            if (-1 == unlink(myaddr.sun_path)){
+            if (-1 == unlink(myaddr.sun_path) && errno != ENOENT){
                 log_perror("unlink failed") << "\t" << myaddr.sun_path << endl;
             }
             old_umask = umask(0);
+            reset_umask = true;
         } else { // Started by user.
             if( getenv( "HOME" )) {
                 string socket_path = getenv("HOME");
@@ -636,7 +645,7 @@ bool Daemon::setup_listen_fds()
                 if(socket_path.length() > sizeof(myaddr.sun_path) - 1) {
                     log_error() << "$HOME/.iceccd.socket path too long for sun_path" << endl;
                 }
-                if (-1 == unlink(myaddr.sun_path)){
+                if (-1 == unlink(myaddr.sun_path) && errno != ENOENT){
                     log_perror("unlink failed") << "\t" << myaddr.sun_path << endl;
                 }
             } else {
@@ -651,7 +660,7 @@ bool Daemon::setup_listen_fds()
         if(test_socket.length() > sizeof(myaddr.sun_path) - 1) {
             log_error() << "$ICECC_TEST_SOCKET path too long for sun_path" << endl;
         }
-        if (-1 == unlink(myaddr.sun_path)){
+        if (-1 == unlink(myaddr.sun_path) && errno != ENOENT){
             log_perror("unlink failed") << "\t" << myaddr.sun_path << endl;
         }
     }
@@ -659,14 +668,14 @@ bool Daemon::setup_listen_fds()
     if (bind(unix_listen_fd, (struct sockaddr*)&myaddr, sizeof(myaddr)) < 0) {
         log_perror("bind()");
 
-        if (old_umask != -1U) {
+        if (reset_umask) {
             umask(old_umask);
         }
 
         return false;
     }
 
-    if (old_umask != -1U) {
+    if (reset_umask) {
         umask(old_umask);
     }
 
@@ -751,24 +760,25 @@ void Daemon::close_scheduler()
     delete discover;
     discover = 0;
     next_scheduler_connect = time(0) + 20 + (rand() & 31);
+    static bool fast_reconnect = getenv( "ICECC_TESTS" ) != NULL;
+    if( fast_reconnect )
+        next_scheduler_connect = time(0) + 3;
 }
 
-bool Daemon::maybe_stats(bool send_ping)
+bool Daemon::maybe_stats(bool force_check)
 {
     struct timeval now;
     gettimeofday(&now, 0);
 
     time_t diff_sent = (now.tv_sec - last_stat.tv_sec) * 1000 + (now.tv_usec - last_stat.tv_usec) / 1000;
 
-    if (diff_sent >= max_scheduler_pong * 1000) {
+    if (diff_sent >= max_scheduler_pong * 1000 || force_check) {
         StatsMsg msg;
         unsigned int memory_fillgrade;
         unsigned long idleLoad = 0;
         unsigned long niceLoad = 0;
 
-        if (!fill_stats(idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes)) {
-            return false;
-        }
+        fill_stats(idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes);
 
         time_t diff_stat = (now.tv_sec - last_stat.tv_sec) * 1000 + (now.tv_usec - last_stat.tv_usec) / 1000;
         last_stat = now;
@@ -795,40 +805,34 @@ bool Daemon::maybe_stats(bool send_ping)
             icecream_usage.tv_usec = ru.ru_utime.tv_usec;
         }
 
-        int idle_average = icecream_load;
+        unsigned int idle_average = icecream_load;
 
         if (diff_sent) {
             idle_average = icecream_load * 1000 / diff_sent;
         }
 
-        if (idle_average > 1000) {
-            idle_average = 1000;
-        }
+        if (idle_average > 1000)
+           idle_average = 1000;
 
-        msg.load = ((700 * (1000 - idle_average)) + (300 * memory_fillgrade)) / 1000;
-
-        if (memory_fillgrade > 600) {
-            msg.load = 1000;
-        }
-
-        if (idle_average < 100) {
-            msg.load = 1000;
-        }
+        msg.load = std::max((1000 - idle_average), memory_fillgrade);
 
 #ifdef HAVE_SYS_VFS_H
         struct statfs buf;
         int ret = statfs(envbasedir.c_str(), &buf);
 
-        if (!ret && long(buf.f_bavail) < ((long(max_kids + 1 - current_kids) * 4 * 1024 * 1024) / buf.f_bsize)) {
+        // Require at least 25MiB of free disk space per build.
+        if (!ret && long(buf.f_bavail) < ((long(max_kids + 1 - current_kids) * 25 * 1024 * 1024) / buf.f_bsize)) {
             msg.load = 1000;
         }
 
 #endif
 
         // Matz got in the urine that not all CPUs are always feed
-        mem_limit = std::max(int(msg.freeMem / std::min(std::max(max_kids, 1U), 4U)), int(100U));
+        mem_limit = std::max(int(msg.freeMem / std::min(std::max(max_kids, 1U), 4U)), min_mem_limit);
 
-        if (abs(int(msg.load) - current_load) >= 100 || send_ping) {
+        if (abs(int(msg.load) - current_load) >= 100
+            || (msg.load == 1000 && current_load != 1000)
+            || (msg.load != 1000 && current_load == 1000)) {
             if (!send_scheduler(msg)) {
                 return false;
             }
@@ -888,14 +892,13 @@ string Daemon::dump_internals() const
     unsigned long idleLoad = 0;
     unsigned long niceLoad = 0;
 
-    if (fill_stats(idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes)) {
-        result += "  cpu: " + toString(idleLoad) + " idle, "
-                  + toString(niceLoad) + " nice\n";
-        result += "  load: " + toString(msg.loadAvg1 / 1000.) + ", icecream_load: "
-                  + toString(icecream_load) + "\n";
-        result += "  memory: " + toString(memory_fillgrade)
-                  + " (free: " + toString(msg.freeMem) + ")\n";
-    }
+    fill_stats(idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes);
+    result += "  cpu: " + toString(idleLoad) + " idle, "
+              + toString(niceLoad) + " nice\n";
+    result += "  load: " + toString(msg.loadAvg1 / 1000.) + ", icecream_load: "
+              + toString(icecream_load) + "\n";
+    result += "  memory: " + toString(memory_fillgrade)
+              + " (free: " + toString(msg.freeMem) + ")\n";
 
     return result;
 }
@@ -913,8 +916,8 @@ int Daemon::scheduler_use_cs(UseCSMsg *msg)
             << " " << c << " " << msg->hostname << " " << remote_name <<  endl;
 
     if (!c) {
-        if (send_scheduler(JobDoneMsg(msg->job_id, 107, JobDoneMsg::FROM_SUBMITTER))) {
-            return 1;
+        if (send_scheduler(JobDoneMsg(msg->job_id, 107, JobDoneMsg::FROM_SUBMITTER, clients.size()))) {
+            return 0;
         }
 
         return 1;
@@ -944,12 +947,12 @@ int Daemon::scheduler_use_cs(UseCSMsg *msg)
 int Daemon::scheduler_no_cs(NoCSMsg *msg)
 {
     Client *c = clients.find_by_client_id(msg->client_id);
-    trace() << "handle_use_cs " << msg->job_id << " " << msg->client_id
+    trace() << "handle_no_cs " << msg->job_id << " " << msg->client_id
             << " " << c << " " <<  endl;
 
     if (!c) {
-        if (send_scheduler(JobDoneMsg(msg->job_id, 107, JobDoneMsg::FROM_SUBMITTER))) {
-            return 1;
+        if (send_scheduler(JobDoneMsg(msg->job_id, 107, JobDoneMsg::FROM_SUBMITTER, clients.size()))) {
+            return 0;
         }
 
         return 1;
@@ -1044,8 +1047,7 @@ bool Daemon::handle_transfer_env_done(Client *client)
 
     bool r = reannounce_environments(); // do that before the file compiles
 
-    // we do that here so we're not given out in case of full discs
-    if (!maybe_stats(true)) {
+    if (!maybe_stats(true)) { // update stats in case our disk is too full to accept more jobs
         r = false;
     }
 
@@ -1228,10 +1230,18 @@ bool Daemon::create_env_finished(string env_key)
     trace() << "cache_size = " << cache_size << endl;
 
     if (!installed_size) {
-        for (Clients::const_iterator it = clients.begin(); it != clients.end(); ++it)  {
-            if (it->second->pending_create_env == env_key) {
-                it->second->channel->send_msg(EndMsg());
-                handle_end(it->second, 121);
+        bool repeat = true;
+        while(repeat) {
+            repeat = false;
+            for (Clients::const_iterator it = clients.begin(); it != clients.end(); ++it)  {
+                if (it->second->pending_create_env == env_key) {
+                    it->second->channel->send_msg(EndMsg());
+                    handle_end(it->second, 121);
+                    // The handle_end call invalidates our iterator, so break out of the loop,
+                    // but try again just in case, until there's no match.
+                    repeat = true;
+                    break;
+                }
             }
         }
         return false;
@@ -1265,6 +1275,9 @@ bool Daemon::handle_job_done(Client *cl, JobDoneMsg *m)
 
     assert(msg->job_id == cl->job_id);
     cl->job_id = 0; // the scheduler doesn't have it anymore
+
+    msg->client_count = clients.size();
+
     return send_scheduler(*msg);
 }
 
@@ -1324,7 +1337,7 @@ void Daemon::handle_old_request()
             int sock = -1;
             pid_t pid = -1;
 
-            trace() << "requests--" << job->jobID() << endl;
+            trace() << "request for job " << job->jobID() << endl;
 
             string envforjob = job->targetPlatform() + "/" + job->environmentVersion();
             envs_last_use[envforjob] = time(NULL);
@@ -1337,7 +1350,7 @@ void Daemon::handle_old_request()
                 client->pipe_to_child = sock;
                 client->child_pid = pid;
 
-                if (!send_scheduler(JobBeginMsg(job->jobID()))) {
+                if (!send_scheduler(JobBeginMsg(job->jobID(), clients.size()))) {
                     log_info() << "failed sending scheduler about " << job->jobID() << endl;
                 }
             } else {
@@ -1357,7 +1370,7 @@ bool Daemon::handle_compile_done(Client *client)
     assert(client->child_pid > 0);
     assert(client->pipe_to_child >= 0);
 
-    JobDoneMsg *msg = new JobDoneMsg(client->job->jobID(), -1, JobDoneMsg::FROM_SERVER);
+    JobDoneMsg *msg = new JobDoneMsg(client->job->jobID(), -1, JobDoneMsg::FROM_SERVER, clients.size());
     assert(msg);
     assert(current_kids > 0);
     current_kids--;
@@ -1374,7 +1387,6 @@ bool Daemon::handle_compile_done(Client *client)
         msg->user_msec = job_stat[JobStatistics::user_msec];
         msg->sys_msec = job_stat[JobStatistics::sys_msec];
         msg->pfaults = job_stat[JobStatistics::sys_pfaults];
-        end_status = job_stat[JobStatistics::exit_code];
     }
 
     close(client->pipe_to_child);
@@ -1398,7 +1410,7 @@ bool Daemon::handle_compile_file(Client *client, Msg *msg)
     if (client->status == Client::CLIENTWORK) {
         assert(job->environmentVersion() == "__client");
 
-        if (!send_scheduler(JobBeginMsg(job->jobID()))) {
+        if (!send_scheduler(JobBeginMsg(job->jobID(), clients.size()))) {
             trace() << "can't reach scheduler to tell him about compile file job "
                     << job->jobID() << endl;
             return false;
@@ -1476,17 +1488,21 @@ void Daemon::handle_end(Client *client, int exitcode)
 
     if (scheduler && client->status != Client::WAITFORCHILD) {
         int job_id = client->job_id;
+        bool use_client_id = false;
 
         if (client->status == Client::TOCOMPILE) {
             job_id = client->job->jobID();
         }
 
         if (client->status == Client::WAITFORCS) {
-            job_id = client->client_id; // it's all we have
-            exitcode = CLIENT_WAS_WAITING_FOR_CS; // this is the message
+            // We don't know the job id, because we haven't received a reply
+            // from the scheduler yet. Use client_id to identify the job,
+            // the scheduler will use it for matching.
+            use_client_id = true;
+            assert( client->client_id > 0 );
         }
 
-        if (job_id > 0) {
+        if (job_id > 0 || use_client_id) {
             JobDoneMsg::from_type flag = JobDoneMsg::FROM_SUBMITTER;
 
             switch (client->status) {
@@ -1512,7 +1528,11 @@ void Daemon::handle_end(Client *client, int exitcode)
 
             trace() << "scheduler->send_msg( JobDoneMsg( " << client->dump() << ", " << exitcode << "))\n";
 
-            if (!send_scheduler(JobDoneMsg(job_id, exitcode, flag))) {
+            JobDoneMsg msg(job_id, exitcode, flag, clients.size());
+            if( use_client_id ) {
+                msg.set_unknown_job_client_id( client->client_id );
+            }
+            if (!send_scheduler(msg)) {
                 trace() << "failed to reach scheduler for remote job done msg!" << endl;
             }
         } else if (client->status == Client::CLIENTWORK) {
@@ -1571,6 +1591,8 @@ bool Daemon::handle_get_cs(Client *client, Msg *msg)
         return true;
     }
 
+    umsg->client_count = clients.size();
+
     return send_scheduler(*umsg);
 }
 
@@ -1643,7 +1665,7 @@ bool Daemon::handle_activity(Client *client)
 {
     assert(client->status != Client::TOCOMPILE);
 
-    Msg *msg = client->channel->get_msg();
+    Msg *msg = client->channel->get_msg(0, true);
 
     if (!msg) {
         handle_end(client, 118);
@@ -1702,7 +1724,7 @@ bool Daemon::handle_activity(Client *client)
     return ret;
 }
 
-int Daemon::answer_client_requests()
+void Daemon::answer_client_requests()
 {
 #ifdef ICECC_DEBUG
 
@@ -1808,7 +1830,14 @@ int Daemon::answer_client_requests()
 
     if (ret < 0 && errno != EINTR) {
         log_perror("select");
-        return 5;
+        close_scheduler();
+        return;
+    }
+    // Reset debug if needed, but only if we aren't waiting for any child processes to finish,
+    // otherwise their debug output could end up reset in the middle (and flush log marks used
+    // by tests could be written out before debug output from children).
+    if( current_kids == 0 ) {
+        reset_debug_if_needed();
     }
 
     if (ret > 0) {
@@ -1816,13 +1845,13 @@ int Daemon::answer_client_requests()
 
         if (scheduler && FD_ISSET(scheduler->fd, &listen_set)) {
             while (!scheduler->read_a_bit() || scheduler->has_msg()) {
-                Msg *msg = scheduler->get_msg();
+                Msg *msg = scheduler->get_msg(0, true);
 
                 if (!msg) {
                     log_error() << "scheduler closed connection" << endl;
                     close_scheduler();
                     clear_children();
-                    return 1;
+                    return;
                 }
 
                 ret = 0;
@@ -1855,7 +1884,8 @@ int Daemon::answer_client_requests()
                 delete msg;
 
                 if (ret) {
-                    return ret;
+                    close_scheduler();
+                    return;
                 }
             }
         }
@@ -1881,13 +1911,13 @@ int Daemon::answer_client_requests()
 
             if (acc_fd == -1 && errno != EINTR) {
                 log_perror("accept failed:");
-                return EXIT_CONNECT_FAILED;
+                return;
             }
 
             MsgChannel *c = Service::createChannel(acc_fd, &cli_addr, cli_len);
 
             if (!c) {
-                return 0;
+                return;
             }
 
             trace() << "accepted " << c->fd << " " << c->name << endl;
@@ -1924,7 +1954,7 @@ int Daemon::answer_client_requests()
                     max_fd--;
 
                     if (!handle_compile_done(client)) {
-                        return 1;
+                        return;
                     }
                 }
 
@@ -1962,12 +1992,10 @@ int Daemon::answer_client_requests()
 
         if (had_scheduler && !scheduler) {
             clear_children();
-            return 2;
+            return;
         }
 
     }
-
-    return 0;
 }
 
 bool Daemon::reconnect()
@@ -1977,7 +2005,7 @@ bool Daemon::reconnect()
     }
 
     if (!discover && next_scheduler_connect > time(0)) {
-        trace() << "timeout.." << endl;
+        trace() << "Delaying reconnect." << endl;
         return false;
     }
 
@@ -1991,7 +2019,7 @@ bool Daemon::reconnect()
     }
 
     if (!scheduler) {
-        log_warning() << "scheduler not yet found." << endl;
+        log_warning() << "scheduler not yet found/selected." << endl;
         return false;
     }
 
@@ -2023,13 +2051,7 @@ int Daemon::working_loop()
 {
     for (;;) {
         reconnect();
-
-        int ret = answer_client_requests();
-
-        if (ret) {
-            trace() << "answer_client_requests returned " << ret << endl;
-            close_scheduler();
-        }
+        answer_client_requests();
 
         if (exit_main_loop) {
             close_scheduler();
@@ -2071,7 +2093,7 @@ int main(int argc, char **argv)
             { 0, 0, 0, 0 }
         };
 
-        const int c = getopt_long(argc, argv, "N:n:m:l:s:whvdrb:u:p:", long_options, &option_index);
+        const int c = getopt_long(argc, argv, "N:n:m:l:s:hvdb:u:p:", long_options, &option_index);
 
         if (c == -1) {
             break;    // eoo
@@ -2138,14 +2160,8 @@ int main(int argc, char **argv)
             break;
         case 'v':
 
-            if (debug_level & Warning)
-                if (debug_level & Info) { // for second call
-                    debug_level |= Debug;
-                } else {
-                    debug_level |= Info;
-                }
-            else {
-                debug_level |= Warning;
+            if (debug_level < MaxVerboseLevel) {
+                debug_level++;
             }
 
             break;
@@ -2233,7 +2249,7 @@ int main(int argc, char **argv)
     }
 
     if (d.warn_icecc_user_errno != 0) {
-        log_errno("Error: no icecc user on system. Falling back to nobody.", d.warn_icecc_user_errno);
+        log_errno("No icecc user on system. Falling back to nobody.", d.warn_icecc_user_errno);
     }
 
     umask(022);
