@@ -52,11 +52,13 @@
 #include "../services/comm.h"
 #include "../services/logging.h"
 #include "../services/job.h"
+#include "../services/util.h"
 #include "config.h"
 
 #include "compileserver.h"
 #include "job.h"
 
+// Values 0 to 3.
 #define DEBUG_SCHEDULER 0
 
 /* TODO:
@@ -120,8 +122,7 @@ static list<UnansweredList *> toanswer;
 static list<JobStat> all_job_stats;
 static JobStat cum_job_stats;
 
-static float server_speed(CompileServer *cs, Job *job = 0);
-static void broadcast_scheduler_version(const char* netname);
+static float server_speed(CompileServer *cs, Job *job = 0, bool blockDebug = false);
 
 /* Searches the queue for JOB and removes it.
    Returns true if something was deleted.  */
@@ -220,7 +221,7 @@ static void add_job_stats(Job *job, JobDoneMsg *msg)
                 << " " << st.outputSize() << " " << msg->out_uncompressed << " "
                 << job->server()->nodeName() << " "
                 << float(msg->out_uncompressed) / st.compileTimeUser() << " "
-                << server_speed(job->server()) << endl;
+                << server_speed(job->server(), NULL, true) << endl;
     }
 #endif
 }
@@ -245,8 +246,11 @@ static void notify_monitors(Msg *m)
     delete m;
 }
 
-static float server_speed(CompileServer *cs, Job *job)
+static float server_speed(CompileServer *cs, Job *job, bool blockDebug)
 {
+#if DEBUG_SCHEDULER <= 2
+    (void)blockDebug;
+#endif
     if (cs->lastCompiledJobs().size() == 0 || cs->cumCompiled().compileTimeUser() == 0) {
         return 0;
     } else {
@@ -256,21 +260,57 @@ static float server_speed(CompileServer *cs, Job *job)
         // we only care for the load if we're about to add a job to it
         if (job) {
             if (job->submitter() == cs) {
-                /* The submitter of a job gets more speed if it's capable of handling its requests on its own.
-                   So if he is equally fast to the rest of the farm it will be preferred to chose him
-                   to compile the job.  Then this can be done locally without needing the preprocessor.
-                   However if there are more requests than the number of jobs the submitter can handle,
-                   it is assumed the submitter is doing a massively parallel build, in which case it is
-                   better not to build on the submitter and let it do other work (such as preprocessing
-                   output for other nodes) that can be done only locally.  */
-                if (cs->submittedJobsCount() <= cs->maxJobs()) {
-                    f *= 1.1;
-                } else {
-                    f *= 0.1;    // penalize heavily
+                int clientCount = cs->clientCount();
+                if( clientCount == 0 ) {
+                    // Older client/daemon that doesn't send client count. Use the number of jobs
+                    // that we've already been told about as the fallback value (it will sometimes
+                    // be an underestimate).
+                    clientCount = cs->submittedJobsCount();
                 }
-            } else { // ignoring load for submitter - assuming the load is our own
+                if (clientCount > cs->maxJobs()) {
+                    // The submitter would be overloaded by building all its jobs locally,
+                    // so penalize it heavily in order to send jobs preferably to other nodes,
+                    // so that the submitter should preferably do tasks that cannot be distributed,
+                    // such as linking or preparing jobs for remote nodes.
+                    f *= 0.1;
+#if DEBUG_SCHEDULER > 2
+                    if(!blockDebug)
+                        log_info() << "penalizing local build for job " << job->id() << endl;
+#endif
+                } else if (clientCount == cs->maxJobs()) {
+                    // This means the submitter would be fully loaded by its jobs. It is still
+                    // preferable to distribute the job, unless the submitter is noticeably faster.
+                    f *= 0.8;
+#if DEBUG_SCHEDULER > 2
+                    if(!blockDebug)
+                        log_info() << "slightly penalizing local build for job " << job->id() << endl;
+#endif
+                }
+                else if (clientCount <= cs->maxJobs() / 2) {
+                    // The submitter has only few jobs, slightly prefer building the job locally
+                    // in order to save the overhead of distributing.
+                    // Note that this is unreliable, the submitter may be in fact running a large
+                    // parallel build but this is just the first of the jobs and other icecc instances
+                    // haven't been launched yet. There's probably no good way to detect this reliably.
+                    f *= 1.1;
+#if DEBUG_SCHEDULER > 2
+                    if(!blockDebug)
+                        log_info() << "slightly preferring local build for job " << job->id() << endl;
+#endif
+                } else {
+                    // the remaining case, don't adjust
+                    f *= 1;
+                }
+                // ignoring load for submitter - assuming the load is our own
+            } else {
                 f *= float(1000 - cs->load()) / 1000;
             }
+
+            /* Gradually throttle with the number of assigned jobs. This
+             * takes care of the fact that not all slots are equally fast on
+             * CPUs with SMT and dynamic clock ramping.
+             */
+            f *= (1.0f - (0.5f * cs->jobList().size() / cs->maxJobs()));
         }
 
         // below we add a pessimism factor - assuming the first job a computer got is not representative
@@ -306,16 +346,16 @@ static void handle_monitor_stats(CompileServer *cs, StatsMsg *m = 0)
     if (m) {
         sprintf(buffer, "Load:%d\n", m->load);
         msg += buffer;
-        sprintf(buffer, "LoadAvg1:%d\n", m->loadAvg1);
+        sprintf(buffer, "LoadAvg1:%u\n", m->loadAvg1);
         msg += buffer;
-        sprintf(buffer, "LoadAvg5:%d\n", m->loadAvg5);
+        sprintf(buffer, "LoadAvg5:%u\n", m->loadAvg5);
         msg += buffer;
-        sprintf(buffer, "LoadAvg10:%d\n", m->loadAvg10);
+        sprintf(buffer, "LoadAvg10:%u\n", m->loadAvg10);
         msg += buffer;
-        sprintf(buffer, "FreeMem:%d\n", m->freeMem);
+        sprintf(buffer, "FreeMem:%u\n", m->freeMem);
         msg += buffer;
     } else {
-        sprintf(buffer, "Load:%d\n", cs->load());
+        sprintf(buffer, "Load:%u\n", cs->load());
         msg += buffer;
     }
 
@@ -385,6 +425,8 @@ static bool handle_cs_request(MsgChannel *cs, Msg *_m)
 
     CompileServer *submitter = static_cast<CompileServer *>(cs);
 
+    submitter->setClientCount(m->client_count);
+
     Job *master_job = 0;
 
     for (unsigned int i = 0; i < m->count; ++i) {
@@ -392,7 +434,26 @@ static bool handle_cs_request(MsgChannel *cs, Msg *_m)
         job->setEnvironments(m->versions);
         job->setTargetPlatform(m->target);
         job->setArgFlags(m->arg_flags);
-        job->setLanguage((m->lang == CompileJob::Lang_C) ? "C" : "C++");
+        switch(m->lang) {
+            case CompileJob::Lang_C:
+                job->setLanguage("C");
+                break;
+            case CompileJob::Lang_CXX:
+                job->setLanguage("C++");
+                break;
+            case CompileJob::Lang_OBJC:
+                job->setLanguage("ObjC");
+                break;
+            case CompileJob::Lang_OBJCXX:
+                job->setLanguage("ObjC++");
+                break;
+            case CompileJob::Lang_Custom:
+                job->setLanguage("<custom>");
+                break;
+            default:
+                job->setLanguage("???"); // presumably newer client?
+                break;
+        }
         job->setFileName(m->filename);
         job->setLocalClientId(m->client_id);
         job->setPreferredHost(m->preferred_host);
@@ -557,18 +618,6 @@ static CompileServer *pick_server(Job *job)
         return 0;
     }
 
-    /* Now guess about the job.  First see, if this submitter already
-       had other jobs.  Use them as base.  */
-    JobStat guess;
-
-    if (job->submitter()->lastRequestedJobs().size() > 0) {
-        guess = job->submitter()->cumRequested()
-                / job->submitter()->lastRequestedJobs().size();
-    } else {
-        /* Otherwise simply average over all jobs.  */
-        guess = cum_job_stats / all_job_stats.size();
-    }
-
     CompileServer *best = 0;
     // best uninstalled
     CompileServer *bestui = 0;
@@ -619,8 +668,9 @@ static CompileServer *pick_server(Job *job)
 
 #if DEBUG_SCHEDULER > 1
         trace() << cs->nodeName() << " compiled " << cs->lastCompiledJobs().size() << " got now: " <<
-                cs->jobList().size() << " speed: " << server_speed(cs, job) << " compile time " <<
-                cs->cumCompiled().compileTimeUser() << " produced code " << cs->cumCompiled().outputSize() << endl;
+                cs->jobList().size() << " speed: " << server_speed(cs, job, true) << " compile time " <<
+                cs->cumCompiled().compileTimeUser() << " produced code " << cs->cumCompiled().outputSize() <<
+                " client count: " << cs->clientCount() << endl;
 #endif
 
         if ((cs->lastCompiledJobs().size() == 0) && (cs->jobList().size() == 0) && cs->maxJobs()) {
@@ -638,6 +688,15 @@ static CompileServer *pick_server(Job *job)
             break;
         }
 
+        /* Distribute 5% of our jobs to servers which haven't been picked in a
+           long time. This gives us a chance to adjust the server speed rating,
+           which may change due to external influences out of our control. */
+        if (!cs->lastPickedId() ||
+            ((job->id() - cs->lastPickedId()) > (20 * css.size()))) {
+            best = cs;
+            break;
+        }
+
         if (!envs_match(cs, job).empty()) {
             if (!best) {
                 best = cs;
@@ -671,29 +730,23 @@ static CompileServer *pick_server(Job *job)
         }
     }
 
-    // to make sure we find the fast computers at least after some time, we overwrite
-    // the install rule for every 19th job - if the farm is only filled a bit
-    if (bestui && ((matches < 11) && (matches < (css.size() / 3))) && ((job->id() % 19) != 0)) {
-        best = 0;
-    }
-
     if (best) {
 #if DEBUG_SCHEDULER > 1
-        trace() << "taking best installed " << best->nodeName() << " " <<  server_speed(best, job) << endl;
+        trace() << "taking best installed " << best->nodeName() << " " <<  server_speed(best, job, true) << endl;
 #endif
         return best;
     }
 
     if (bestui) {
 #if DEBUG_SCHEDULER > 1
-        trace() << "taking best uninstalled " << bestui->nodeName() << " " <<  server_speed(bestui, job) << endl;
+        trace() << "taking best uninstalled " << bestui->nodeName() << " " <<  server_speed(bestui, job, true) << endl;
 #endif
         return bestui;
     }
 
     if (bestpre) {
 #if DEBUG_SCHEDULER > 1
-        trace() << "taking best preload " << bestui->nodeName() << " " <<  server_speed(bestui, job) << endl;
+        trace() << "taking best preload " << bestui->nodeName() << " " <<  server_speed(bestui, job, true) << endl;
 #endif
     }
 
@@ -1070,6 +1123,8 @@ static bool handle_job_begin(CompileServer *cs, Msg *_m)
         return false;
     }
 
+    cs->setClientCount(m->client_count);
+
     job->setState(Job::COMPILING);
     job->setStartTime(m->stime);
     job->setStartOnScheduler(time(0));
@@ -1095,21 +1150,21 @@ static bool handle_job_done(CompileServer *cs, Msg *_m)
 
     Job *j = 0;
 
-    if (m->exitcode == CLIENT_WAS_WAITING_FOR_CS) {
-        // the daemon saw a cancel of what he believes is waiting in the scheduler
+    if (uint32_t clientId = m->unknown_job_client_id()) {
+        // The daemon has sent a done message for a job for which it doesn't know the job id (happens
+        // if the job is cancelled before we send back the job id). Find the job using the client id.
         map<unsigned int, Job *>::iterator mit;
 
         for (mit = jobs.begin(); mit != jobs.end(); ++mit) {
             Job *job = mit->second;
             trace() << "looking for waitcs " << job->server() << " " << job->submitter()  << " " << cs
-                    << " " << job->state() << " " << job->localClientId() << " " << m->job_id
+                    << " " << job->state() << " " << job->localClientId() << " " << clientId
                     << endl;
 
-            if (job->server() == 0 && job->submitter() == cs && job->state() == Job::PENDING
-                    && job->localClientId() == m->job_id) {
+            if (job->server() == 0 && job->submitter() == cs && job->localClientId() == clientId) {
                 trace() << "STOP (WAITFORCS) FOR " << mit->first << endl;
                 j = job;
-                m->job_id = j->id(); // that's faked
+                m->set_job_id( j->id()); // Now we know the job's id.
 
                 /* Unfortunately the toanswer queues are also tagged based on the daemon,
                 so we need to clean them up also.  */
@@ -1143,11 +1198,6 @@ static bool handle_job_done(CompileServer *cs, Msg *_m)
         return false;
     }
 
-    if (j->state() == Job::PENDING) {
-        trace() << "job ID still pending ?! scheduler recently restarted? " << m->job_id << endl;
-        return false;
-    }
-
     if (m->is_from_server() && (j->server() != cs)) {
         log_info() << "the server isn't the same for job " << m->job_id << endl;
         log_info() << "server: " << j->server()->nodeName() << endl;
@@ -1166,6 +1216,8 @@ static bool handle_job_done(CompileServer *cs, Msg *_m)
         return false;
     }
 
+    cs->setClientCount(m->client_count);
+
     if (m->exitcode == 0) {
         std::ostream &dbg = trace();
         dbg << "END " << m->job_id
@@ -1240,6 +1292,7 @@ static bool handle_stats(CompileServer *cs, Msg *_m)
     for (list<CompileServer *>::iterator it = css.begin(); it != css.end(); ++it)
         if (*it == cs) {
             (*it)->setLoad(m->load);
+            (*it)->setClientCount(m->client_count);
             handle_monitor_stats(*it, m);
             return true;
         }
@@ -1284,7 +1337,7 @@ static string dump_job(Job *job)
     default:
         jobState = "Huh?";
     }
-    snprintf(buffer, sizeof(buffer), "%d %s sub:%s on:%s ",
+    snprintf(buffer, sizeof(buffer), "%u %s sub:%s on:%s ",
              job->id(),
              jobState.c_str(),
              job->submitter() ? job->submitter()->nodeName().c_str() : "<>",
@@ -1347,7 +1400,6 @@ static bool handle_line(CompileServer *cs, Msg *_m)
         return false;
     }
 
-    char buffer[1000];
     string line;
     list<string> l;
     split_string(m->text, " \t\n", l);
@@ -1365,6 +1417,7 @@ static bool handle_line(CompileServer *cs, Msg *_m)
 
     if (cmd == "listcs") {
         for (list<CompileServer *>::iterator it = css.begin(); it != css.end(); ++it) {
+            char buffer[1000];
             sprintf(buffer, " (%s:%d) ", (*it)->name.c_str(), (*it)->remotePort());
             line = " " + (*it)->nodeName() + buffer;
             line += "[" + (*it)->hostPlatform() + "] speed=";
@@ -1622,7 +1675,7 @@ static bool handle_activity(CompileServer *cs)
 {
     Msg *m;
     bool ret = true;
-    m = cs->get_msg(0);
+    m = cs->get_msg(0, true);
 
     if (!m) {
         handle_end(cs, m);
@@ -1751,48 +1804,6 @@ static int open_tcp_listener(short port)
     return fd;
 }
 
-#define BROAD_BUFLEN 268
-#define BROAD_BUFLEN_OLD 16
-static int prepare_broadcast_reply(char* buf, const char* netname)
-{
-    if (buf[0] < 33) { // old client
-        buf[0]++;
-        memset(buf + 1, 0, BROAD_BUFLEN_OLD - 1);
-        snprintf(buf + 1, BROAD_BUFLEN_OLD - 1, "%s", netname);
-        buf[BROAD_BUFLEN_OLD - 1] = 0;
-        return BROAD_BUFLEN_OLD;
-    } else { // net client
-        buf[0] += 2;
-        memset(buf + 1, 0, BROAD_BUFLEN - 1);
-        uint32_t tmp_version = PROTOCOL_VERSION;
-        uint64_t tmp_time = starttime;
-        memcpy(buf + 1, &tmp_version, sizeof(uint32_t));
-        memcpy(buf + 1 + sizeof(uint32_t), &tmp_time, sizeof(uint64_t));
-        const int OFFSET = 1 + sizeof(uint32_t) + sizeof(uint64_t);
-        snprintf(buf + OFFSET, BROAD_BUFLEN - OFFSET, "%s", netname);
-        buf[BROAD_BUFLEN - 1] = 0;
-        return BROAD_BUFLEN;
-    }
-}
-
-static void broadcast_scheduler_version(const char* netname)
-{
-    const char length_netname = strlen(netname);
-    const int schedbuflen = 5 + sizeof(uint64_t) + length_netname;
-    char *buf = new char[ schedbuflen ];
-    buf[0] = 'I';
-    buf[1] = 'C';
-    buf[2] = 'E';
-    buf[3] = PROTOCOL_VERSION;
-    uint64_t tmp_time = starttime;
-    memcpy(buf + 4, &tmp_time, sizeof(uint64_t));
-    buf[4 + sizeof(uint64_t)] = length_netname;
-    strncpy(buf + 5 + sizeof(uint64_t), netname, length_netname);
-    DiscoverSched::broadcastData(scheduler_port, buf, schedbuflen);
-    delete[] buf;
-    buf = 0;
-}
-
 static void usage(const char *reason = 0)
 {
     if (reason) {
@@ -1822,7 +1833,7 @@ static void trigger_exit(int signum)
     } else {
         // hmm, we got killed already. try better
         static const char msg[] = "forced exit.\n";
-        write(STDERR_FILENO, msg, strlen( msg ));
+        ignore_result(write(STDERR_FILENO, msg, strlen( msg )));
         _exit(1);
     }
 
@@ -1834,26 +1845,26 @@ static void handle_scheduler_announce(const char* buf, const char* netname, bool
 {
     /* Another scheduler is announcing it's running, disconnect daemons if it has a better version
        or the same version but was started earlier. */
-    if (!persistent_clients){
-        uint64_t tmp_time;
-        memcpy(&tmp_time, buf + 4, sizeof(uint64_t));
-        time_t other_time = tmp_time;
-        const unsigned char other_scheduler_protocol = buf[3];
-        if (other_scheduler_protocol >= 36)
+    time_t other_time;
+    int other_protocol_version;
+    string other_netname;
+    Broadcasts::getSchedulerVersionData(buf, &other_protocol_version, &other_time, &other_netname);
+    trace() << "Received scheduler announcement from " << inet_ntoa(broad_addr.sin_addr)
+            << ":" << ntohs(broad_addr.sin_port)
+            << " (version " << int(other_protocol_version) << ", netname " << other_netname << ")" << endl;
+    if (other_protocol_version >= 36)
+    {
+        if (other_netname == netname)
         {
-            const unsigned char recv_netname_len = buf[4 + sizeof(uint64_t)];
-            string local_netname = netname;
-            string recv_netname = string(buf + 5 + sizeof(uint64_t), recv_netname_len);
-            if (recv_netname == local_netname)
+            if (other_protocol_version > PROTOCOL_VERSION || (other_protocol_version == PROTOCOL_VERSION && other_time < starttime))
             {
-                if (other_scheduler_protocol > PROTOCOL_VERSION || (other_scheduler_protocol == PROTOCOL_VERSION && other_time < starttime))
-                {
+                if (!persistent_clients){
+                    log_info() << "Scheduler from " << inet_ntoa(broad_addr.sin_addr)
+                        << ":" << ntohs(broad_addr.sin_port)
+                        << " (version " << int(other_protocol_version) << ") has announced itself as a preferred"
+                        " scheduler, disconnecting all connections." << endl;
                     if (!css.empty() || !monitors.empty())
                     {
-                        log_info() << "Scheduler from " << inet_ntoa(broad_addr.sin_addr)
-                            << ":" << ntohs(broad_addr.sin_port)
-                            << " (version " << int(other_scheduler_protocol) << ") has announced itself as a preferred"
-                            " scheduler, disconnecting all connections." << endl;
                         while (!css.empty())
                         {
                             handle_end(css.front(), NULL);
@@ -1874,7 +1885,7 @@ int main(int argc, char *argv[])
     int listen_fd, remote_fd, broad_fd, text_fd;
     struct sockaddr_in remote_addr;
     socklen_t remote_len;
-    char *netname = (char *)"ICECREAM";
+    const char *netname = "ICECREAM";
     bool detach = false;
     bool persistent_clients = false;
     int debug_level = Error;
@@ -1912,7 +1923,7 @@ int main(int argc, char *argv[])
             { 0, 0, 0, 0 }
         };
 
-        const int c = getopt_long(argc, argv, "n:p:hl:vdr:u:r:", long_options, &option_index);
+        const int c = getopt_long(argc, argv, "n:p:hl:vdru:", long_options, &option_index);
 
         if (c == -1) {
             break;    // eoo
@@ -1938,14 +1949,8 @@ int main(int argc, char *argv[])
             break;
         case 'v':
 
-            if (debug_level & Warning) {
-                if (debug_level & Info) { // for second call
-                    debug_level |= Debug;
-                } else {
-                    debug_level |= Info;
-                }
-            } else {
-                debug_level |= Warning;
+            if (debug_level < MaxVerboseLevel) {
+                debug_level++;
             }
 
             break;
@@ -1999,7 +2004,7 @@ int main(int argc, char *argv[])
     }
 
     if (warn_icecc_user_errno != 0) {
-        log_errno("Error: no icecc user on system. Falling back to nobody.", errno);
+        log_errno("No icecc user on system. Falling back to nobody.", errno);
     }
 
     if (getuid() == 0) {
@@ -2040,7 +2045,10 @@ int main(int argc, char *argv[])
     log_info() << "ICECREAM scheduler " VERSION " starting up, port " << scheduler_port << endl;
 
     if (detach) {
-        daemon(0, 0);
+        if (daemon(0, 0) != 0) {
+            log_errno("Failed to detach.", errno);
+            exit(1);
+        }
     }
 
     listen_fd = open_tcp_listener(scheduler_port);
@@ -2067,6 +2075,8 @@ int main(int argc, char *argv[])
     }
 
     starttime = time(0);
+    if( getenv( "ICECC_FAKE_STARTTIME" ) != NULL )
+        starttime -= 1000;
 
     ofstream pidFile;
     string progName = argv[0];
@@ -2080,9 +2090,11 @@ int main(int argc, char *argv[])
     signal(SIGINT, trigger_exit);
     signal(SIGALRM, trigger_exit);
 
+    log_info() << "scheduler ready" << endl;
+
     time_t next_listen = 0;
 
-    broadcast_scheduler_version(netname);
+    Broadcasts::broadcastSchedulerVersion(scheduler_port, netname, starttime);
     last_announce = starttime;
 
     while (!exit_main_loop) {
@@ -2098,7 +2110,7 @@ int main(int argc, char *argv[])
            their daemons if we are the preferred scheduler (daemons with version new enough
            should automatically select the best scheduler, but old daemons connect randomly). */
         if (last_announce + 120 < time(NULL)) {
-            broadcast_scheduler_version(netname);
+            Broadcasts::broadcastSchedulerVersion(scheduler_port, netname, starttime);
             last_announce = time(NULL);
         }
 
@@ -2163,19 +2175,21 @@ int main(int argc, char *argv[])
             }
         }
 
-        max_fd = select(max_fd + 1, &read_set, &write_set, NULL, &tv);
+        int active_fds = select(max_fd + 1, &read_set, &write_set, NULL, &tv);
 
-        if (max_fd < 0 && errno == EINTR) {
+        if (active_fds < 0 && errno == EINTR) {
+            reset_debug_if_needed(); // we possibly got SIGHUP
             continue;
         }
+        reset_debug_if_needed();
 
-        if (max_fd < 0) {
+        if (active_fds < 0) {
             log_perror("select()");
             return 1;
         }
 
         if (FD_ISSET(listen_fd, &read_set)) {
-            max_fd--;
+            active_fds--;
             bool pending_connections = true;
 
             while (pending_connections) {
@@ -2218,8 +2232,8 @@ int main(int argc, char *argv[])
             next_listen = time(0) + 1;
         }
 
-        if (max_fd && FD_ISSET(text_fd, &read_set)) {
-            max_fd--;
+        if (active_fds && FD_ISSET(text_fd, &read_set)) {
+            active_fds--;
             remote_len = sizeof(remote_addr);
             remote_fd = accept(text_fd,
                                (struct sockaddr *) &remote_addr,
@@ -2247,20 +2261,17 @@ int main(int argc, char *argv[])
             }
         }
 
-        if (max_fd && FD_ISSET(broad_fd, &read_set)) {
-            max_fd--;
-            char buf[BROAD_BUFLEN + 1];
+        if (active_fds && FD_ISSET(broad_fd, &read_set)) {
+            active_fds--;
+            char buf[Broadcasts::BROAD_BUFLEN + 1];
             struct sockaddr_in broad_addr;
             socklen_t broad_len = sizeof(broad_addr);
             /* We can get either a daemon request for a scheduler (1 byte) or another scheduler
                announcing itself (4 bytes + time). */
 
-            int schedbuflen = 4 + sizeof(uint64_t);
-
-            int buflen = recvfrom(broad_fd, buf, BROAD_BUFLEN, 0, (struct sockaddr *) &broad_addr,
+            int buflen = recvfrom(broad_fd, buf, Broadcasts::BROAD_BUFLEN, 0, (struct sockaddr *) &broad_addr,
                     &broad_len);
-            /* Daemon is searching for a scheduler, only answer if daemon would be able to talk to us. */
-            if (buflen < 0 || buflen > BROAD_BUFLEN){
+            if (buflen < 0 || buflen > Broadcasts::BROAD_BUFLEN){
                 int err = errno;
                 log_perror("recvfrom()");
 
@@ -2273,42 +2284,27 @@ int main(int argc, char *argv[])
                     return -1;
                 }
             }
-            if (buflen == 1) {
-                if (buf[0] >= MIN_PROTOCOL_VERSION){
+            int daemon_version;
+            if (DiscoverSched::isSchedulerDiscovery(buf, buflen, &daemon_version)) {
+                /* Daemon is searching for a scheduler, only answer if daemon would be able to talk to us. */
+                if ( daemon_version >= MIN_PROTOCOL_VERSION){
                     log_info() << "broadcast from " << inet_ntoa(broad_addr.sin_addr)
                         << ":" << ntohs(broad_addr.sin_port)
-                        << " (version " << int(buf[0]) << ")\n";
-                    int reply_len = prepare_broadcast_reply(buf, netname);
+                        << " (version " << daemon_version << ")\n";
+                    int reply_len = DiscoverSched::prepareBroadcastReply(buf, netname, starttime);
                     if (sendto(broad_fd, buf, reply_len, 0,
                                 (struct sockaddr *) &broad_addr, broad_len) != reply_len) {
                         log_perror("sendto()");
                     }
                 }
             }
-            else if(buflen >= schedbuflen && buf[0] == 'I' && buf[1] == 'C' && buf[2] == 'E') {
-                if(buf[3] > 35)
-                {
-                    schedbuflen += 1 + buf[schedbuflen];
-                }
-                if (buflen != schedbuflen){
-                    int err = errno;
-                    log_perror("recvfrom()");
-
-                    /* Some linux 2.6 kernels can return from select with
-                       data available, and then return from read() with EAGAIN
-                       even on a blocking socket (breaking POSIX).  Happens
-                       when the arriving packet has a wrong checksum.  So
-                       we ignore EAGAIN here, but still abort for all other errors. */
-                    if (err != EAGAIN) {
-                        return -1;
-                    }
-                }
+            else if(Broadcasts::isSchedulerVersion(buf, buflen)) {
                 handle_scheduler_announce(buf, netname, persistent_clients, broad_addr);
             }
         }
 
         for (map<int, CompileServer *>::const_iterator it = fd2cs.begin();
-                max_fd && it != fd2cs.end();) {
+                active_fds > 0 && it != fd2cs.end();) {
             int i = it->first;
             CompileServer *cs = it->second;
             /* handle_activity can delete the channel from the fd2cs list,
@@ -2323,20 +2319,23 @@ int main(int argc, char *argv[])
                     }
                 }
 
-                max_fd--;
+                active_fds--;
             }
         }
 
         for (list<CompileServer *>::const_iterator it = cs_in_tsts.begin();
                 it != cs_in_tsts.end(); ++it) {
+            if(find(css.begin(), css.end(), *it) == css.end()) {
+                continue; // deleted meanwhile
+            }
             if((*it)->getConnectionInProgress())
             {
-                if(max_fd && (FD_ISSET((*it)->getInFd(), &read_set) || FD_ISSET((*it)->getInFd(), &write_set)) && (*it)->isConnected())
+                if(active_fds > 0 && (FD_ISSET((*it)->getInFd(), &read_set) || FD_ISSET((*it)->getInFd(), &write_set)) && (*it)->isConnected())
                 {
-                    max_fd--;
+                    active_fds--;
                     (*it)->updateInConnectivity(true);
                 }
-                else if((!max_fd || (FD_ISSET((*it)->getInFd(), &read_set) || FD_ISSET((*it)->getInFd(), &write_set))) && !(*it)->isConnected())
+                else if((active_fds == 0 || (FD_ISSET((*it)->getInFd(), &read_set) || FD_ISSET((*it)->getInFd(), &write_set))) && !(*it)->isConnected())
                 {
                     (*it)->updateInConnectivity(false);
                 }
@@ -2352,7 +2351,7 @@ int main(int argc, char *argv[])
     if ((-1 == close(broad_fd)) && (errno != EBADF)){
         log_perror("close failed");
     }
-    if (-1 == unlink(pidFilePath.c_str())){
+    if (-1 == unlink(pidFilePath.c_str()) && errno != ENOENT){
         log_perror("unlink failed") << "\t" << pidFilePath << endl;
     }
     return 0;
