@@ -50,6 +50,9 @@
 #ifdef HAVE_LIBCAP_NG
 #include <cap-ng.h>
 #endif
+#include "getifaddrs.h"
+#include <net/if.h>
+#include <sys/ioctl.h>
 
 #include "logging.h"
 #include "job.h"
@@ -66,6 +69,14 @@ using namespace std;
 
 #define MAX_MSG_SIZE 1 * 1024 * 1024
 
+/*
+ * On a slow and congested network it's possible for a send call to get starved.
+ * This will happen especially when trying to send a huge number of bytes over at
+ * once. We can avoid this situation to a large extend by sending smaller
+ * chunks of data over.
+ */
+#define MAX_WRITE_SIZE 10 * 1024
+
 /* TODO
  * buffered in/output per MsgChannel
     + move read* into MsgChannel, create buffer-fill function
@@ -87,6 +98,7 @@ bool MsgChannel::read_a_bit()
     if (count < 128) {
         inbuflen = (inbuflen + 128 + 127) & ~(size_t) 127;
         inbuf = (char *) realloc(inbuf, inbuflen);
+        assert(inbuf); // Probably unrecoverable if realloc fails anyway.
         count = inbuflen - inofs;
     }
 
@@ -123,7 +135,14 @@ bool MsgChannel::read_a_bit()
         error = true;
     }
 
-    return !error;
+    if (error) {
+        // Daemons sometimes successfully do accept() but then the connection
+        // gets ECONNRESET. Probably a spurious result from accept(), so
+        // just be silent about it in this case.
+        set_error( instate == NEED_PROTO );
+        return false;
+    }
+    return true;
 }
 
 bool MsgChannel::update_state(void)
@@ -152,6 +171,7 @@ bool MsgChannel::update_state(void)
 
                 if (remote_prot < MIN_PROTOCOL_VERSION || remote_prot > (1 << 20)) {
                     remote_prot = 0;
+                    set_error();
                     return false;
                 }
 
@@ -166,6 +186,7 @@ bool MsgChannel::update_state(void)
                 writefull(vers, 4);
 
                 if (!flush_writebuf(true)) {
+                    set_error();
                     return false;
                 }
 
@@ -176,6 +197,7 @@ bool MsgChannel::update_state(void)
 
                 if ((int)remote_prot != protocol) {
                     protocol = 0;
+                    set_error();
                     return false;
                 }
 
@@ -184,6 +206,8 @@ bool MsgChannel::update_state(void)
                 break;
             } else {
                 trace() << "NEED_PROTO but protocol > 0" << endl;
+                set_error();
+                return false;
             }
         }
 
@@ -192,7 +216,7 @@ bool MsgChannel::update_state(void)
         if (instate != NEED_LEN) {
             break;
         }
-
+        // fallthrough
     case NEED_LEN:
 
         if (text_based) {
@@ -214,12 +238,15 @@ bool MsgChannel::update_state(void)
             (*this) >> inmsglen;
 
             if (inmsglen > MAX_MSG_SIZE) {
+                log_error() << "received a too large message (size " << inmsglen << "), ignoring" << endl;
+                set_error();
                 return false;
             }
 
             if (inbuflen - intogo < inmsglen) {
                 inbuflen = (inmsglen + intogo + 127) & ~(size_t)127;
                 inbuf = (char *) realloc(inbuf, inbuflen);
+                assert(inbuf); // Probably unrecoverable if realloc fails anyway.
             }
 
             instate = FILL_BUF;
@@ -227,7 +254,7 @@ bool MsgChannel::update_state(void)
         } else {
             break;
         }
-
+        /* FALLTHROUGH */
     case FILL_BUF:
 
         if (inofs - intogo >= inmsglen) {
@@ -241,6 +268,9 @@ bool MsgChannel::update_state(void)
     case HAS_MSG:
         /* handled elsewere */
         break;
+
+    case ERROR:
+        return false;
     }
 
     return true;
@@ -277,6 +307,7 @@ void MsgChannel::writefull(const void *_buf, size_t count)
         /* Realloc to a multiple of 128.  */
         msgbuflen = (msgtogo + count + 127) & ~(size_t)127;
         msgbuf = (char *) realloc(msgbuf, msgbuflen);
+        assert(msgbuf); // Probably unrecoverable if realloc fails anyway.
     }
 
     memcpy(msgbuf + msgtogo, _buf, count);
@@ -290,12 +321,12 @@ bool MsgChannel::flush_writebuf(bool blocking)
 
     while (msgtogo) {
 #ifdef MSG_NOSIGNAL
-        ssize_t ret = send(fd, buf, msgtogo, MSG_NOSIGNAL);
+        ssize_t ret = send(fd, buf, msgtogo < MAX_WRITE_SIZE ? msgtogo : MAX_WRITE_SIZE, MSG_NOSIGNAL);
 #else
         void (*oldsigpipe)(int);
 
         oldsigpipe = signal(SIGPIPE, SIG_IGN);
-        ssize_t ret = send(fd, buf, msgtogo, 0);
+        ssize_t ret = send(fd, buf, msgtogo < MAX_WRITE_SIZE ? msgtogo : MAX_WRITE_SIZE, 0);
         signal(SIGPIPE, oldsigpipe);
 #endif
 
@@ -306,7 +337,7 @@ bool MsgChannel::flush_writebuf(bool blocking)
 
             /* If we want to write blocking, but couldn't write anything,
                select on the fd.  */
-            if (blocking && errno == EAGAIN) {
+            if (blocking && ( errno == EAGAIN || errno == ENOTCONN )) {
                 int ready;
 
                 for (;;) {
@@ -348,7 +379,11 @@ bool MsgChannel::flush_writebuf(bool blocking)
 
     msgofs = buf - msgbuf;
     chop_output();
-    return !error;
+    if(error) {
+        set_error();
+        return false;
+    }
+    return true;
 }
 
 MsgChannel &MsgChannel::operator>>(uint32_t &buf)
@@ -482,6 +517,7 @@ void MsgChannel::readcompressed(unsigned char **uncompressed_buf, size_t &_uclen
         uncompressed_len = 0;
         _uclen = uncompressed_len;
         _clen = compressed_len;
+        set_error();
         return;
     }
 
@@ -528,6 +564,7 @@ void MsgChannel::writecompressed(const unsigned char *in_buf, size_t _in_len, si
         /* Realloc to a multiple of 128.  */
         msgbuflen = (msgtogo + out_len + 127) & ~(size_t)127;
         msgbuf = (char *) realloc(msgbuf, msgbuflen);
+        assert(msgbuf); // Probably unrecoverable if realloc fails anyway.
     }
 
     lzo_byte *out_buf = (lzo_byte *)(msgbuf + msgtogo);
@@ -542,6 +579,9 @@ void MsgChannel::writecompressed(const unsigned char *in_buf, size_t _in_len, si
     }
 
     uint32_t _olen = htonl(out_len);
+    if(out_len > MAX_MSG_SIZE) {
+        log_error() << "internal error - size of compressed message to write exceeds max size:" << out_len << endl;
+    }
     memcpy(msgbuf + msgtogo_old, &_olen, 4);
     msgtogo += out_len;
     _out_len = out_len;
@@ -573,6 +613,18 @@ void MsgChannel::write_line(const string &line)
     }
 }
 
+void MsgChannel::set_error(bool silent)
+{
+    if( instate == ERROR ) {
+        return;
+    }
+    if( !silent ) {
+        trace() << "setting error state for channel " << dump() << endl;
+    }
+    instate = ERROR;
+    eof = true;
+}
+
 static int prepare_connect(const string &hostname, unsigned short p,
                            struct sockaddr_in &remote_addr)
 {
@@ -692,10 +744,11 @@ MsgChannel *Service::createChannel(const string &hostname, unsigned short p, int
         setsockopt(remote_fd, SOL_SOCKET, SO_SNDBUF, &i, sizeof(i));
 
         if (connect(remote_fd, (struct sockaddr *) &remote_addr, sizeof(remote_addr)) < 0) {
+            log_perror_trace("connect");
+            trace() << "connect failed on " << hostname << endl;
             if (-1 == close(remote_fd) && (errno != EBADF)){
                 log_perror("close failed");
             }
-            trace() << "connect failed on " << hostname << endl;
             return 0;
         }
     }
@@ -722,10 +775,11 @@ MsgChannel *Service::createChannel(const string &socket_path)
     }
 
     if (connect(remote_fd, (struct sockaddr *) &remote_addr, sizeof(remote_addr)) < 0) {
+        log_perror_trace("connect");
+        trace() << "connect failed on " << socket_path << endl;
         if ((-1 == close(remote_fd)) && (errno != EBADF)){
             log_perror("close failed");
         }
-        trace() << "connect failed on " << socket_path << endl;
         return 0;
     }
 
@@ -803,6 +857,7 @@ MsgChannel::MsgChannel(int _fd, struct sockaddr *_a, socklen_t _l, bool text)
     int on = 1;
 
     if (!setsockopt(_fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &on, sizeof(on))) {
+#if defined( TCP_KEEPIDLE ) || defined( TCPCTL_KEEPIDLE )
 #if defined( TCP_KEEPIDLE )
         int keepidle = TCP_KEEPIDLE;
 #else
@@ -812,7 +867,9 @@ MsgChannel::MsgChannel(int _fd, struct sockaddr *_a, socklen_t _l, bool text)
         int sec;
         sec = MAX_SCHEDULER_PING - 3 * MAX_SCHEDULER_PONG;
         setsockopt(_fd, IPPROTO_TCP, keepidle, (char *) &sec, sizeof(sec));
+#endif
 
+#if defined( TCP_KEEPINTVL ) || defined( TCPCTL_KEEPINTVL )
 #if defined( TCP_KEEPINTVL )
         int keepintvl = TCP_KEEPINTVL;
 #else
@@ -821,6 +878,7 @@ MsgChannel::MsgChannel(int _fd, struct sockaddr *_a, socklen_t _l, bool text)
 
         sec = MAX_SCHEDULER_PONG;
         setsockopt(_fd, IPPROTO_TCP, keepintvl, (char *) &sec, sizeof(sec));
+#endif
 
 #ifdef TCP_KEEPCNT
         sec = 3;
@@ -848,6 +906,7 @@ MsgChannel::MsgChannel(int _fd, struct sockaddr *_a, socklen_t _l, bool text)
 
         if (!flush_writebuf(true)) {
             protocol = 0;    // unusable
+            set_error();
         }
     }
 
@@ -887,7 +946,7 @@ string MsgChannel::dump() const
 bool MsgChannel::wait_for_protocol()
 {
     /* protocol is 0 if we couldn't send our initial protocol version.  */
-    if (protocol == 0) {
+    if (protocol == 0 || instate == ERROR) {
         return false;
     }
 
@@ -905,12 +964,14 @@ bool MsgChannel::wait_for_protocol()
         }
 
         if (ret == 0) {
-            log_error() << "no response from local daemon within timeout." << endl;
+            log_error() << "no response within timeout" << endl;
+            set_error();
             return false; /* timeout. Consider it a fatal error. */
         }
 
         if (ret < 0) {
             log_perror("select in wait_for_protocol()");
+            set_error();
             return false;
         }
 
@@ -944,17 +1005,22 @@ void MsgChannel::setBulkTransfer()
    message to arrive.  Returns false if there was some error.  */
 bool MsgChannel::wait_for_msg(int timeout)
 {
+    if (instate == ERROR) {
+        return false;
+    }
+
     if (has_msg()) {
         return true;
     }
 
     if (!read_a_bit()) {
         trace() << "!read_a_bit\n";
+        set_error();
         return false;
     }
 
     if (timeout <= 0) {
-        trace() << "timeout <= 0\n";
+        // trace() << "timeout <= 0\n";
         return has_msg();
     }
 
@@ -978,6 +1044,7 @@ bool MsgChannel::wait_for_msg(int timeout)
 
         if (!read_a_bit()) {
             trace() << "!read_a_bit 2\n";
+            set_error();
             return false;
         }
     }
@@ -985,14 +1052,13 @@ bool MsgChannel::wait_for_msg(int timeout)
     return true;
 }
 
-Msg *MsgChannel::get_msg(int timeout)
+Msg *MsgChannel::get_msg(int timeout, bool eofAllowed)
 {
     Msg *m = 0;
     enum MsgType type;
-    uint32_t t;
 
     if (!wait_for_msg(timeout)) {
-        trace() << "!wait_for_msg()\n";
+        // trace() << "!wait_for_msg()\n";
         return 0;
     }
 
@@ -1000,24 +1066,32 @@ Msg *MsgChannel::get_msg(int timeout)
        then we won't see it anymore.  Return that to the caller.
        Don't use has_msg() here, as it returns true for eof.  */
     if (at_eof()) {
-        trace() << "saw eof without complete msg! " << instate << endl;
+        if (!eofAllowed) {
+            trace() << "saw eof without complete msg! " << instate << endl;
+            set_error();
+        }
         return 0;
     }
 
     if (!has_msg()) {
         trace() << "saw eof without msg! " << eof << " " << instate << endl;
+        set_error();
         return 0;
     }
 
+    size_t intogo_old = intogo;
+
     if (text_based) {
         type = M_TEXT;
     } else {
+        uint32_t t;
         *this >> t;
         type = (enum MsgType) t;
     }
 
     switch (type) {
     case M_UNKNOWN:
+        set_error();
         return 0;
     case M_PING:
         m = new PingMsg;
@@ -1115,10 +1189,22 @@ Msg *MsgChannel::get_msg(int timeout)
 
     if (!m) {
         trace() << "no message type" << endl;
+        set_error();
         return 0;
     }
 
     m->fill_from_channel(this);
+
+    if (!text_based) {
+        if( intogo - intogo_old != inmsglen ) {
+            log_error() << "internal error - message not read correctly, message size " << inmsglen
+                << " read " << (intogo - intogo_old) << endl;
+            delete m;
+            set_error();
+            return 0;
+        }
+    }
+
     instate = NEED_LEN;
     update_state();
 
@@ -1127,6 +1213,9 @@ Msg *MsgChannel::get_msg(int timeout)
 
 bool MsgChannel::send_msg(const Msg &m, int flags)
 {
+    if (instate == ERROR) {
+        return false;
+    }
     if (instate == NEED_PROTO && !wait_for_protocol()) {
         return false;
     }
@@ -1139,7 +1228,13 @@ bool MsgChannel::send_msg(const Msg &m, int flags)
     } else {
         *this << (uint32_t) 0;
         m.send_to_channel(this);
-        uint32_t len = htonl(msgtogo - msgtogo_old - 4);
+        uint32_t out_len = msgtogo - msgtogo_old - 4;
+        if(out_len > MAX_MSG_SIZE) {
+            log_error() << "internal error - size of message to write exceeds max size:" << out_len << endl;
+            set_error();
+            return false;
+        }
+        uint32_t len = htonl(out_len);
         memcpy(msgbuf + msgtogo_old, &len, 4);
     }
 
@@ -1150,9 +1245,97 @@ bool MsgChannel::send_msg(const Msg &m, int flags)
     return flush_writebuf((flags & SendBlocking));
 }
 
-#include "getifaddrs.h"
-#include <net/if.h>
-#include <sys/ioctl.h>
+static int get_second_port_for_debug( int port )
+{
+    // When running tests, we want to check also interactions between 2 schedulers, but
+    // when they are both local, they cannot bind to the same port. So make sure to
+    // send all broadcasts to both.
+    static bool checkedDebug = false;
+    static int debugPort1 = 0;
+    static int debugPort2 = 0;
+    if( !checkedDebug ) {
+        checkedDebug = true;
+        if( const char* env = getenv( "ICECC_TEST_SCHEDULER_PORTS" )) {
+            debugPort1 = atoi( env );
+            const char* env2 = strchr( env, ':' );
+            if( env2 != NULL )
+                debugPort2 = atoi( env2 + 1 );
+        }
+    }
+    int secondPort = 0;
+    if( port == debugPort1 )
+        secondPort = debugPort2;
+    else if( port == debugPort2 )
+        secondPort = debugPort1;
+    return secondPort ? secondPort : -1;
+}
+
+void Broadcasts::broadcastSchedulerVersion(int scheduler_port, const char* netname, time_t starttime)
+{
+    // Code for older schedulers than version 38. Has endianness problems, the message size
+    // is not BROAD_BUFLEN and the netname is possibly not null-terminated.
+    const char length_netname = strlen(netname);
+    const int schedbuflen = 5 + sizeof(uint64_t) + length_netname;
+    char *buf = new char[ schedbuflen ];
+    buf[0] = 'I';
+    buf[1] = 'C';
+    buf[2] = 'E';
+    buf[3] = PROTOCOL_VERSION;
+    uint64_t tmp_time = starttime;
+    memcpy(buf + 4, &tmp_time, sizeof(uint64_t));
+    buf[4 + sizeof(uint64_t)] = length_netname;
+    strncpy(buf + 5 + sizeof(uint64_t), netname, length_netname);
+    broadcastData(scheduler_port, buf, schedbuflen);
+    delete[] buf;
+    // Latest version.
+    buf = new char[ BROAD_BUFLEN ];
+    memset(buf, 0, BROAD_BUFLEN );
+    buf[0] = 'I';
+    buf[1] = 'C';
+    buf[2] = 'F'; // one up
+    buf[3] = PROTOCOL_VERSION;
+    uint32_t tmp_time_low = starttime & 0xffffffffUL;
+    uint32_t tmp_time_high = uint64_t(starttime) >> 32;
+    tmp_time_low = htonl( tmp_time_low );
+    tmp_time_high = htonl( tmp_time_high );
+    memcpy(buf + 4, &tmp_time_high, sizeof(uint32_t));
+    memcpy(buf + 4 + sizeof(uint32_t), &tmp_time_low, sizeof(uint32_t));
+    const int OFFSET = 4 + 2 * sizeof(uint32_t);
+    snprintf(buf + OFFSET, BROAD_BUFLEN - OFFSET, "%s", netname);
+    buf[BROAD_BUFLEN - 1] = 0;
+    broadcastData(scheduler_port, buf, BROAD_BUFLEN);
+    delete[] buf;
+}
+
+bool Broadcasts::isSchedulerVersion(const char* buf, int buflen)
+{
+    if( buflen != BROAD_BUFLEN )
+        return false;
+    // Ignore versions older than 38, they are older than us anyway, so not interesting.
+    if( buf[0] == 'I' && buf[1] == 'C' && buf[2] == 'F') {
+        return true;
+    }
+    return false;
+}
+
+void Broadcasts::getSchedulerVersionData( const char* buf, int* protocol, time_t* time, string* netname )
+{
+    assert( isSchedulerVersion( buf, BROAD_BUFLEN ));
+    const unsigned char other_scheduler_protocol = buf[3];
+    uint32_t tmp_time_low, tmp_time_high;
+    memcpy(&tmp_time_high, buf + 4, sizeof(uint32_t));
+    memcpy(&tmp_time_low, buf + 4 + sizeof(uint32_t), sizeof(uint32_t));
+    tmp_time_low = ntohl( tmp_time_low );
+    tmp_time_high = ntohl( tmp_time_high );
+    time_t other_time = ( uint64_t( tmp_time_high ) << 32 ) | tmp_time_low;;
+    string recv_netname = string(buf + 4 + 2 * sizeof(uint32_t));
+    if( protocol != NULL )
+        *protocol = other_scheduler_protocol;
+    if( time != NULL )
+        *time = other_time;
+    if( netname != NULL )
+        *netname = recv_netname;
+}
 
 /* Returns a filedesc. or a negative value for errors.  */
 static int open_send_broadcast(int port, const char* buf, int size)
@@ -1201,14 +1384,22 @@ static int open_send_broadcast(int port, const char* buf, int size)
             continue;
         }
 
-        if (ntohl(((struct sockaddr_in *) addr->ifa_addr)->sin_addr.s_addr) == 0x7f000001) {
-            trace() << "ignoring localhost " << addr->ifa_name << endl;
-            continue;
-        }
+        static bool in_tests = getenv( "ICECC_TESTS" ) != NULL;
+        if (!in_tests) {
+            if (ntohl(((struct sockaddr_in *) addr->ifa_addr)->sin_addr.s_addr) == 0x7f000001) {
+                trace() << "ignoring localhost " << addr->ifa_name << endl;
+                continue;
+            }
 
-        if ((addr->ifa_flags & IFF_POINTOPOINT) || !(addr->ifa_flags & IFF_BROADCAST)) {
-            log_info() << "ignoring tunnels " << addr->ifa_name << endl;
-            continue;
+            if ((addr->ifa_flags & IFF_POINTOPOINT) || !(addr->ifa_flags & IFF_BROADCAST)) {
+                log_info() << "ignoring tunnels " << addr->ifa_name << endl;
+                continue;
+            }
+        } else {
+            if (ntohl(((struct sockaddr_in *) addr->ifa_addr)->sin_addr.s_addr) != 0x7f000001) {
+                trace() << "ignoring non-localhost " << addr->ifa_name << endl;
+                continue;
+            }
         }
 
         if (addr->ifa_broadaddr) {
@@ -1232,74 +1423,22 @@ static int open_send_broadcast(int port, const char* buf, int size)
     return ask_fd;
 }
 
-#define BROAD_BUFLEN 32
-#define BROAD_BUFLEN_OLD 16
-
-static bool
-get_broad_answer(int ask_fd, int timeout, char *buf2, struct sockaddr_in *remote_addr,
-                 socklen_t *remote_len)
+void Broadcasts::broadcastData(int port, const char* buf, int len)
 {
-    char buf = PROTOCOL_VERSION;
-    fd_set read_set;
-    FD_ZERO(&read_set);
-    FD_SET(ask_fd, &read_set);
-    struct timeval tv;
-    tv.tv_sec = timeout / 1000;
-    tv.tv_usec = 1000 * (timeout % 1000);
-    errno = 0;
-
-    if (select(ask_fd + 1, &read_set, NULL, NULL, &tv) <= 0) {
-        /* Normally this is a timeout, i.e. no scheduler there.  */
-        if (errno) {
-            log_perror("waiting for scheduler");
+    int fd = open_send_broadcast(port, buf, len);
+    if (fd >= 0) {
+        if ((-1 == close(fd)) && (errno != EBADF)){
+            log_perror("close failed");
         }
-
-        return false;
     }
-
-    *remote_len = sizeof(struct sockaddr_in);
-
-    int len = recvfrom(ask_fd, buf2, BROAD_BUFLEN, 0, (struct sockaddr *) remote_addr, remote_len);
-    if (len != BROAD_BUFLEN && len != BROAD_BUFLEN_OLD) {
-        log_perror("get_broad_answer recvfrom()");
-        return false;
-    }
-
-    if ((len == BROAD_BUFLEN_OLD && buf2[0] != buf + 1) // PROTOCOL <= 32 scheduler
-        || (len == BROAD_BUFLEN && buf2[0] != buf + 2)) { // PROTOCOL >= 33 scheduler
-        log_error() << "wrong answer" << endl;
-        return false;
-    }
-
-    buf2[len - 1] = 0;
-    return true;
-}
-
-static void get_broad_data(const char* buf, const char** name, int* version, time_t* start_time)
-{
-    if (buf[0] == PROTOCOL_VERSION + 1) {
-        // Scheduler version 32 or older, didn't send us its version, assume it's 32.
-        if (name != NULL)
-            *name = buf + 1;
-        if (version != NULL)
-            *version = 32;
-        if (start_time != NULL)
-            *start_time = 0; // Unknown too.
-    } else if(buf[0] == PROTOCOL_VERSION + 2) {
-        if (version != NULL) {
-            uint32_t tmp_version;
-            memcpy(&tmp_version, buf + 1, sizeof(uint32_t));
-            *version = tmp_version;
+    int secondPort = get_second_port_for_debug( port );
+    if( secondPort > 0 ) {
+        int fd2 = open_send_broadcast(secondPort, buf, len);
+        if (fd2 >= 0) {
+            if ((-1 == close(fd2)) && (errno != EBADF)){
+                log_perror("close failed");
+            }
         }
-        if (start_time != NULL) {
-            uint64_t tmp_time;
-            memcpy(&tmp_time, buf + 1 + sizeof(uint32_t), sizeof(uint64_t));
-            *start_time = tmp_time;
-        }
-        if (name != NULL)
-            *name = buf + 1 + sizeof(uint32_t) + sizeof(uint64_t);
-    } else {
-        abort();
     }
 }
 
@@ -1309,15 +1448,19 @@ DiscoverSched::DiscoverSched(const std::string &_netname, int _timeout,
     , schedname(_schedname)
     , timeout(_timeout)
     , ask_fd(-1)
+    , ask_second_fd(-1)
     , sport(port)
     , best_version(0)
     , best_start_time(0)
+    , best_port(0)
     , multiple(false)
 {
     time0 = time(0);
 
     if (schedname.empty()) {
-        const char *get = getenv("USE_SCHEDULER");
+        const char *get = getenv("ICECC_SCHEDULER");
+        if( get == NULL )
+            get = getenv("USE_SCHEDULER");
 
         if (get) {
             string scheduler = get;
@@ -1342,8 +1485,7 @@ DiscoverSched::DiscoverSched(const std::string &_netname, int _timeout,
         netname = ""; // take whatever the machine is giving us
         attempt_scheduler_connect();
     } else {
-        char buf = PROTOCOL_VERSION;
-        ask_fd = open_send_broadcast(sport, &buf, 1);
+        sendSchedulerDiscovery( PROTOCOL_VERSION );
     }
 }
 
@@ -1354,6 +1496,11 @@ DiscoverSched::~DiscoverSched()
             log_perror("close failed");
         }
     }
+    if (ask_second_fd >= 0) {
+        if ((-1 == close(ask_second_fd)) && (errno != EBADF)){
+            log_perror("close failed");
+        }
+    }
 }
 
 bool DiscoverSched::timed_out()
@@ -1363,7 +1510,6 @@ bool DiscoverSched::timed_out()
 
 void DiscoverSched::attempt_scheduler_connect()
 {
-
     time0 = time(0) + MAX_SCHEDULER_PONG;
     log_info() << "scheduler is on " << schedname << ":" << sport << " (net " << netname << ")" << endl;
 
@@ -1372,9 +1518,130 @@ void DiscoverSched::attempt_scheduler_connect()
     }
 }
 
+void DiscoverSched::sendSchedulerDiscovery( int version )
+{
+        assert( version < 128 );
+        char buf = version;
+        ask_fd = open_send_broadcast(sport, &buf, 1);
+        int secondPort = get_second_port_for_debug( sport );
+        if( secondPort > 0 )
+            ask_second_fd = open_send_broadcast(secondPort, &buf, 1);
+}
+
+bool DiscoverSched::isSchedulerDiscovery(const char* buf, int buflen, int* daemon_version)
+{
+    if( buflen != 1 )
+        return false;
+    if( daemon_version != NULL ) {
+        *daemon_version = buf[ 0 ];
+    }
+    return true;
+}
+
+static const int BROAD_BUFLEN = 268;
+static const int BROAD_BUFLEN_OLD_2 = 32;
+static const int BROAD_BUFLEN_OLD_1 = 16;
+
+int DiscoverSched::prepareBroadcastReply(char* buf, const char* netname, time_t starttime)
+{
+    if (buf[0] < 33) { // old client
+        buf[0]++;
+        memset(buf + 1, 0, BROAD_BUFLEN_OLD_1 - 1);
+        snprintf(buf + 1, BROAD_BUFLEN_OLD_1 - 1, "%s", netname);
+        buf[BROAD_BUFLEN_OLD_1 - 1] = 0;
+        return BROAD_BUFLEN_OLD_1;
+    } else if (buf[0] < 36) {
+        // This is like 36, but 36 silently changed the size of BROAD_BUFLEN from 32 to 268.
+        // Since get_broad_answer() explicitly null-terminates the data, this wouldn't lead
+        // to those receivers reading a shorter string that would not be null-terminated,
+        // but still, this is what versions 33-35 actually worked with.
+        buf[0] += 2;
+        memset(buf + 1, 0, BROAD_BUFLEN_OLD_2 - 1);
+        uint32_t tmp_version = PROTOCOL_VERSION;
+        uint64_t tmp_time = starttime;
+        memcpy(buf + 1, &tmp_version, sizeof(uint32_t));
+        memcpy(buf + 1 + sizeof(uint32_t), &tmp_time, sizeof(uint64_t));
+        const int OFFSET = 1 + sizeof(uint32_t) + sizeof(uint64_t);
+        snprintf(buf + OFFSET, BROAD_BUFLEN_OLD_2 - OFFSET, "%s", netname);
+        buf[BROAD_BUFLEN_OLD_2 - 1] = 0;
+        return BROAD_BUFLEN_OLD_2;
+    } else if (buf[0] < 38) { // exposes endianess because of not using htonl()
+        buf[0] += 2;
+        memset(buf + 1, 0, BROAD_BUFLEN - 1);
+        uint32_t tmp_version = PROTOCOL_VERSION;
+        uint64_t tmp_time = starttime;
+        memcpy(buf + 1, &tmp_version, sizeof(uint32_t));
+        memcpy(buf + 1 + sizeof(uint32_t), &tmp_time, sizeof(uint64_t));
+        const int OFFSET = 1 + sizeof(uint32_t) + sizeof(uint64_t);
+        snprintf(buf + OFFSET, BROAD_BUFLEN - OFFSET, "%s", netname);
+        buf[BROAD_BUFLEN - 1] = 0;
+        return BROAD_BUFLEN;
+    } else { // latest version
+        buf[0] += 3;
+        memset(buf + 1, 0, BROAD_BUFLEN - 1);
+        uint32_t tmp_version = PROTOCOL_VERSION;
+        uint32_t tmp_time_low = starttime & 0xffffffffUL;
+        uint32_t tmp_time_high = uint64_t(starttime) >> 32;
+        tmp_version = htonl( tmp_version );
+        tmp_time_low = htonl( tmp_time_low );
+        tmp_time_high = htonl( tmp_time_high );
+        memcpy(buf + 1, &tmp_version, sizeof(uint32_t));
+        memcpy(buf + 1 + sizeof(uint32_t), &tmp_time_high, sizeof(uint32_t));
+        memcpy(buf + 1 + 2 * sizeof(uint32_t), &tmp_time_low, sizeof(uint32_t));
+        const int OFFSET = 1 + 3 * sizeof(uint32_t);
+        snprintf(buf + OFFSET, BROAD_BUFLEN - OFFSET, "%s", netname);
+        buf[BROAD_BUFLEN - 1] = 0;
+        return BROAD_BUFLEN;
+    }
+}
+
+void DiscoverSched::get_broad_data(const char* buf, const char** name, int* version, time_t* start_time)
+{
+    if (buf[0] == PROTOCOL_VERSION + 1) {
+        // Scheduler version 32 or older, didn't send us its version, assume it's 32.
+        if (name != NULL)
+            *name = buf + 1;
+        if (version != NULL)
+            *version = 32;
+        if (start_time != NULL)
+            *start_time = 0; // Unknown too.
+    } else if(buf[0] == PROTOCOL_VERSION + 2) {
+        if (version != NULL) {
+            uint32_t tmp_version;
+            memcpy(&tmp_version, buf + 1, sizeof(uint32_t));
+            *version = tmp_version;
+        }
+        if (start_time != NULL) {
+            uint64_t tmp_time;
+            memcpy(&tmp_time, buf + 1 + sizeof(uint32_t), sizeof(uint64_t));
+            *start_time = tmp_time;
+        }
+        if (name != NULL)
+            *name = buf + 1 + sizeof(uint32_t) + sizeof(uint64_t);
+    } else if(buf[0] == PROTOCOL_VERSION + 3) {
+        if (version != NULL) {
+            uint32_t tmp_version;
+            memcpy(&tmp_version, buf + 1, sizeof(uint32_t));
+            *version = ntohl( tmp_version );
+        }
+        if (start_time != NULL) {
+            uint32_t tmp_time_low, tmp_time_high;
+            memcpy(&tmp_time_high, buf + 1 + sizeof(uint32_t), sizeof(uint32_t));
+            memcpy(&tmp_time_low, buf + 1 + 2 * sizeof(uint32_t), sizeof(uint32_t));
+            tmp_time_low = ntohl( tmp_time_low );
+            tmp_time_high = ntohl( tmp_time_high );
+            *start_time = ( uint64_t( tmp_time_high ) << 32 ) | tmp_time_low;;
+        }
+        if (name != NULL)
+            *name = buf + 1 + 3 * sizeof(uint32_t);
+    } else {
+        abort();
+    }
+}
+
 MsgChannel *DiscoverSched::try_get_scheduler()
 {
-    if (schedname.empty() || 0 != best_version) {
+    if (schedname.empty()) {
         socklen_t remote_len;
         char buf2[BROAD_BUFLEN];
         /* Try to get the scheduler with the newest version, and if there
@@ -1392,14 +1659,20 @@ MsgChannel *DiscoverSched::try_get_scheduler()
         */
 
         /* Read/test all packages arrived until now.  */
-        while (get_broad_answer(ask_fd, 0/*timeout*/, buf2,
-                                (struct sockaddr_in *) &remote_addr, &remote_len)) {
+        while (get_broad_answer(ask_fd, 0/*timeout*/, buf2, (struct sockaddr_in *) &remote_addr, &remote_len)
+                || ( ask_second_fd != -1 && get_broad_answer(ask_second_fd, 0/*timeout*/, buf2,
+                                            (struct sockaddr_in *) &remote_addr, &remote_len))) {
             int version;
             time_t start_time;
             const char* name;
             get_broad_data(buf2, &name, &version, &start_time);
             if (strcasecmp(netname.c_str(), name) == 0) {
-                if (version < 33) {
+                if( version >= 128 || version < 1 ) {
+                    log_warning() << "Ignoring bogus version " << version << " from scheduler found at " << inet_ntoa(remote_addr.sin_addr)
+                        << ":" << ntohs(remote_addr.sin_port) << endl;
+                    continue;
+                }
+                else if (version < 33) {
                     log_info() << "Suitable scheduler found at " << inet_ntoa(remote_addr.sin_addr)
                         << ":" << ntohs(remote_addr.sin_port) << " (unknown version)" << endl;
                 } else {
@@ -1409,11 +1682,15 @@ MsgChannel *DiscoverSched::try_get_scheduler()
                 if (best_version != 0)
                     multiple = true;
                 if (best_version < version || (best_version == version && best_start_time > start_time)) {
-                    schedname = inet_ntoa(remote_addr.sin_addr);
-                    sport = ntohs(remote_addr.sin_port);
+                    best_schedname = inet_ntoa(remote_addr.sin_addr);
+                    best_port = ntohs(remote_addr.sin_port);
                     best_version = version;
                     best_start_time = start_time;
                 }
+            } else {
+                log_info() << "Ignoring scheduler at " << inet_ntoa(remote_addr.sin_addr)
+                    << ":" << ntohs(remote_addr.sin_port) << " because of a different netname ("
+                    << name << ")" << endl;
             }
         }
 
@@ -1421,6 +1698,8 @@ MsgChannel *DiscoverSched::try_get_scheduler()
             if (best_version == 0) {
                 return 0;
             }
+            schedname = best_schedname;
+            sport = best_port;
             if (multiple)
                 log_info() << "Selecting scheduler at " << schedname << ":" << sport << endl;
 
@@ -1428,6 +1707,14 @@ MsgChannel *DiscoverSched::try_get_scheduler()
                 log_perror("close failed");
             }
             ask_fd = -1;
+            if( get_second_port_for_debug( sport ) > 0 ) {
+                if (-1 == close(ask_second_fd)){
+                    log_perror("close failed");
+                }
+                ask_second_fd = -1;
+            } else {
+                assert( ask_second_fd == -1 );
+            }
             attempt_scheduler_connect();
 
             if (ask_fd >= 0) {
@@ -1443,6 +1730,7 @@ MsgChannel *DiscoverSched::try_get_scheduler()
         }
     }
     else if (ask_fd >= 0) {
+        assert( ask_second_fd == -1 );
         int status = connect(ask_fd, (struct sockaddr *) &remote_addr, sizeof(remote_addr));
 
         if (status == 0 || (status < 0 && errno == EISCONN)) {
@@ -1456,19 +1744,49 @@ MsgChannel *DiscoverSched::try_get_scheduler()
     return 0;
 }
 
-bool DiscoverSched::broadcastData(int port, const char* buf, int len)
+bool DiscoverSched::get_broad_answer(int ask_fd, int timeout, char *buf2, struct sockaddr_in *remote_addr,
+                 socklen_t *remote_len)
 {
-    int fd = open_send_broadcast(port, buf, len);
-    if (fd >= 0) {
-        if ((-1 == close(fd)) && (errno != EBADF)){
-            log_perror("close failed");
+    char buf = PROTOCOL_VERSION;
+    fd_set read_set;
+    FD_ZERO(&read_set);
+    assert(ask_fd > 0);
+    FD_SET(ask_fd, &read_set);
+    struct timeval tv;
+    tv.tv_sec = timeout / 1000;
+    tv.tv_usec = 1000 * (timeout % 1000);
+    errno = 0;
+
+    if (select(ask_fd + 1, &read_set, NULL, NULL, &tv) <= 0) {
+        /* Normally this is a timeout, i.e. no scheduler there.  */
+        if (errno && errno != EINTR) {
+            log_perror("waiting for scheduler");
         }
-        return true;
+
+        return false;
     }
-    return false;
+
+    *remote_len = sizeof(struct sockaddr_in);
+
+    int len = recvfrom(ask_fd, buf2, BROAD_BUFLEN, 0, (struct sockaddr *) remote_addr, remote_len);
+    if (len != BROAD_BUFLEN && len != BROAD_BUFLEN_OLD_1 && len != BROAD_BUFLEN_OLD_2) {
+        log_perror("get_broad_answer recvfrom()");
+        return false;
+    }
+
+    if (! ((len == BROAD_BUFLEN_OLD_1 && buf2[0] == buf + 1)   // PROTOCOL <= 32 scheduler
+          || (len == BROAD_BUFLEN_OLD_2 && buf2[0] == buf + 2) // PROTOCOL >= 33 && < 36 scheduler
+          || (len == BROAD_BUFLEN && buf2[0] == buf + 2)       // PROTOCOL >= 36 && < 38 scheduler
+          || (len == BROAD_BUFLEN && buf2[0] == buf + 3))) {   // PROTOCOL >= 38 scheduler
+        log_error() << "Wrong scheduler discovery answer (size " << len << ", mark " << int(buf2[0]) << ")" << endl;
+        return false;
+    }
+
+    buf2[len - 1] = 0;
+    return true;
 }
 
-list<string> get_netnames(int timeout, int port)
+list<string> DiscoverSched::getNetnames(int timeout, int port)
 {
     list<string> l;
     int ask_fd;
@@ -1503,6 +1821,11 @@ list<string> get_netnames(int timeout, int port)
     return l;
 }
 
+list<string> get_netnames(int timeout, int port)
+{
+    return DiscoverSched::getNetnames(timeout, port);
+}
+
 void Msg::fill_from_channel(MsgChannel *)
 {
 }
@@ -1548,6 +1871,10 @@ void GetCSMsg::fill_from_channel(MsgChannel *c)
         *c >> version;
         minimal_host_version = max( minimal_host_version, int( version ));
     }
+
+    if (IS_PROTOCOL_39(c)) {
+        *c >> client_count;
+    }
 }
 
 void GetCSMsg::send_to_channel(MsgChannel *c) const
@@ -1571,6 +1898,10 @@ void GetCSMsg::send_to_channel(MsgChannel *c) const
     if (IS_PROTOCOL_34(c)) {
         *c << minimal_host_version;
     }
+
+    if (IS_PROTOCOL_39(c)) {
+        *c << client_count;
+    }
 }
 
 void UseCSMsg::fill_from_channel(MsgChannel *c)
@@ -1790,6 +2121,9 @@ void JobBeginMsg::fill_from_channel(MsgChannel *c)
     Msg::fill_from_channel(c);
     *c >> job_id;
     *c >> stime;
+    if (IS_PROTOCOL_39(c)) {
+        *c >> client_count;
+    }
 }
 
 void JobBeginMsg::send_to_channel(MsgChannel *c) const
@@ -1797,6 +2131,9 @@ void JobBeginMsg::send_to_channel(MsgChannel *c) const
     Msg::send_to_channel(c);
     *c << job_id;
     *c << stime;
+    if (IS_PROTOCOL_39(c)) {
+        *c << client_count;
+    }
 }
 
 void JobLocalBeginMsg::fill_from_channel(MsgChannel *c)
@@ -1827,11 +2164,12 @@ void JobLocalDoneMsg::send_to_channel(MsgChannel *c) const
     *c << job_id;
 }
 
-JobDoneMsg::JobDoneMsg(int id, int exit, unsigned int _flags)
+JobDoneMsg::JobDoneMsg(int id, int exit, unsigned int _flags, unsigned int _client_count)
     : Msg(M_JOB_DONE)
     , exitcode(exit)
     , flags(_flags)
     , job_id(id)
+    , client_count(_client_count)
 {
     real_msec = 0;
     user_msec = 0;
@@ -1859,13 +2197,25 @@ void JobDoneMsg::fill_from_channel(MsgChannel *c)
     *c >> out_uncompressed;
     *c >> flags;
     exitcode = (int) _exitcode;
+    // Older versions used this special exit code to identify
+    // EndJob messages for jobs with unknown job id.
+    if (!IS_PROTOCOL_39(c) && exitcode == 200) {
+        flags |= UnknownJobId;
+    }
+    if (IS_PROTOCOL_39(c)) {
+        *c >> client_count;
+    }
 }
 
 void JobDoneMsg::send_to_channel(MsgChannel *c) const
 {
     Msg::send_to_channel(c);
     *c << job_id;
-    *c << (uint32_t) exitcode;
+    if (!IS_PROTOCOL_39(c) && (flags & UnknownJobId)) {
+        *c << (uint32_t) 200;
+    } else {
+        *c << (uint32_t) exitcode;
+    }
     *c << real_msec;
     *c << user_msec;
     *c << sys_msec;
@@ -1875,9 +2225,32 @@ void JobDoneMsg::send_to_channel(MsgChannel *c) const
     *c << out_compressed;
     *c << out_uncompressed;
     *c << flags;
+    if (IS_PROTOCOL_39(c)) {
+        *c << client_count;
+    }
 }
 
-LoginMsg::LoginMsg(unsigned int myport, const std::string &_nodename, const std::string _host_platform)
+void JobDoneMsg::set_unknown_job_client_id( uint32_t clientId )
+{
+    flags |= UnknownJobId;
+    job_id = clientId;
+}
+
+uint32_t JobDoneMsg::unknown_job_client_id() const
+{
+    if( flags & UnknownJobId ) {
+        return job_id;
+    }
+    return 0;
+}
+
+void JobDoneMsg::set_job_id( uint32_t jobId )
+{
+    job_id = jobId;
+    flags &= ~ (uint32_t) UnknownJobId;
+}
+
+LoginMsg::LoginMsg(unsigned int myport, const std::string &_nodename, const std::string &_host_platform)
     : Msg(M_LOGIN)
     , port(myport)
     , max_kids(0)
