@@ -36,7 +36,7 @@
 #include "job.h"
 
 // if you increase the PROTOCOL_VERSION, add a macro below and use that
-#define PROTOCOL_VERSION 37
+#define PROTOCOL_VERSION 39
 // if you increase the MIN_PROTOCOL_VERSION, comment out macros below and clean up the code
 #define MIN_PROTOCOL_VERSION 21
 
@@ -62,6 +62,13 @@
 #define IS_PROTOCOL_35(c) ((c)->protocol >= 35)
 #define IS_PROTOCOL_36(c) ((c)->protocol >= 36)
 #define IS_PROTOCOL_37(c) ((c)->protocol >= 37)
+#define IS_PROTOCOL_38(c) ((c)->protocol >= 38)
+#define IS_PROTOCOL_39(c) ((c)->protocol >= 39)
+
+// Terms used:
+// S  = scheduler
+// C  = client
+// CS = daemon
 
 enum MsgType {
     // so far unknown
@@ -125,6 +132,7 @@ enum MsgType {
 
     // C --> CS, after installing an environment
     M_VERIFY_ENV,
+    // CS --> C
     M_VERIFY_ENV_RESULT,
     // C --> CS, CS --> S (forwarded from C), to not use given host for given environment
     M_BLACKLIST_HOST_ENV,
@@ -165,7 +173,8 @@ public:
 
     std::string dump() const;
     // NULL  <--> channel closed or timeout
-    Msg *get_msg(int timeout = 10);
+    // Will warn in log if EOF and !eofAllowed.
+    Msg *get_msg(int timeout = 10, bool eofAllowed = false);
 
     // false <--> error (msg not send)
     bool send_msg(const Msg &, int SendFlags = SendBlocking);
@@ -175,6 +184,7 @@ public:
         return eof || instate == HAS_MSG;
     }
 
+    // Returns ture if there were no errors filling inbuf.
     bool read_a_bit(void);
 
     bool at_eof(void) const
@@ -226,6 +236,7 @@ protected:
     void chop_input(void);
     void chop_output(void);
     bool wait_for_msg(int timeout);
+    void set_error(bool silent = false);
 
     char *msgbuf;
     size_t msgbuflen;
@@ -240,7 +251,8 @@ protected:
         NEED_PROTO,
         NEED_LEN,
         FILL_BUF,
-        HAS_MSG
+        HAS_MSG,
+        ERROR
     } instate;
 
     uint32_t inmsglen;
@@ -264,6 +276,21 @@ public:
     static MsgChannel *createChannel(int remote_fd, struct sockaddr *, socklen_t);
 };
 
+class Broadcasts
+{
+public:
+    // Broadcasts a message about this scheduler and its information.
+    static void broadcastSchedulerVersion(int scheduler_port, const char* netname, time_t starttime);
+    // Checks if the data received is a scheduler version broadcast.
+    static bool isSchedulerVersion(const char* buf, int buflen);
+    // Reads data from a scheduler version broadcast.
+    static void getSchedulerVersionData( const char* buf, int* protocol, time_t* time, std::string* netname );
+    /// Broadcasts the given data on the given port.
+    static const int BROAD_BUFLEN = 268;
+private:
+    static void broadcastData(int port, const char* buf, int size);
+};
+
 // --------------------------------------------------------------------------
 // this class is also used by icecream-monitor
 class DiscoverSched
@@ -316,8 +343,15 @@ public:
         return netname;
     }
 
-    /// Broadcasts the given data on the given port.
-    static bool broadcastData(int port, const char* buf, int size);
+    /* Return a list of all reachable netnames.  We wait max. WAITTIME
+       milliseconds for answers.  */
+    static std::list<std::string> getNetnames(int waittime = 2000, int port = 8765);
+
+    // Checks if the data is from a scheduler discovery broadcast, returns version of the sending
+    // daemon is yes.
+    static bool isSchedulerDiscovery(const char* buf, int buflen, int* daemon_version);
+    // Prepares data for sending a reply to a scheduler discovery broadcast.
+    static int prepareBroadcastReply(char* buf, const char* netname, time_t starttime);
 
 private:
     struct sockaddr_in remote_addr;
@@ -325,13 +359,20 @@ private:
     std::string schedname;
     int timeout;
     int ask_fd;
+    int ask_second_fd; // for debugging
     time_t time0;
     unsigned int sport;
     int best_version;
     time_t best_start_time;
+    std::string best_schedname;
+    int best_port;
     bool multiple;
 
     void attempt_scheduler_connect();
+    void sendSchedulerDiscovery( int version );
+    static bool get_broad_answer(int ask_fd, int timeout, char *buf2, struct sockaddr_in *remote_addr,
+                 socklen_t *remote_len);
+    static void get_broad_data(const char* buf, const char** name, int* version, time_t* start_time);
 };
 // --------------------------------------------------------------------------
 
@@ -360,12 +401,14 @@ public:
         : Msg(M_GET_CS)
         , count(1)
         , arg_flags(0)
-        , client_id(0) {}
+        , client_id(0)
+        , client_count(0) {}
 
     GetCSMsg(const Environments &envs, const std::string &f,
              CompileJob::Language _lang, unsigned int _count,
              std::string _target, unsigned int _arg_flags,
-             const std::string &host, int _minimal_host_version)
+             const std::string &host, int _minimal_host_version,
+             unsigned int _client_count = 0)
         : Msg(M_GET_CS)
         , versions(envs)
         , filename(f)
@@ -375,7 +418,8 @@ public:
         , arg_flags(_arg_flags)
         , client_id(0)
         , preferred_host(host)
-        , minimal_host_version(_minimal_host_version) {}
+        , minimal_host_version(_minimal_host_version)
+        , client_count(_client_count) {}
 
     virtual void fill_from_channel(MsgChannel *c);
     virtual void send_to_channel(MsgChannel *c) const;
@@ -389,6 +433,7 @@ public:
     uint32_t client_id;
     std::string preferred_host;
     int minimal_host_version;
+    uint32_t client_count; // number of CS -> C connections at the moment
 };
 
 class UseCSMsg : public Msg
@@ -549,22 +594,21 @@ class JobBeginMsg : public Msg
 {
 public:
     JobBeginMsg()
-        : Msg(M_JOB_BEGIN) {}
+        : Msg(M_JOB_BEGIN)
+        , client_count(0) {}
 
-    JobBeginMsg(unsigned int j)
+    JobBeginMsg(unsigned int j, unsigned int _client_count)
         : Msg(M_JOB_BEGIN)
         , job_id(j)
-        , stime(time(0)) {}
+        , stime(time(0))
+        , client_count(_client_count) {}
 
     virtual void fill_from_channel(MsgChannel *c);
     virtual void send_to_channel(MsgChannel *c) const;
 
     uint32_t job_id;
     uint32_t stime;
-};
-
-enum SpecialExits {
-    CLIENT_WAS_WAITING_FOR_CS = 200
+    uint32_t client_count; // number of CS -> C connections at the moment
 };
 
 class JobDoneMsg : public Msg
@@ -579,7 +623,13 @@ public:
         FROM_SUBMITTER = 1
     };
 
-    JobDoneMsg(int job_id = 0, int exitcode = -1, unsigned int flags = FROM_SERVER);
+    // other flags
+    enum {
+        UnknownJobId = (1 << 1)
+    };
+
+    JobDoneMsg(int job_id = 0, int exitcode = -1, unsigned int flags = FROM_SERVER,
+               unsigned int _client_count = 0);
 
     void set_from(from_type from)
     {
@@ -591,6 +641,10 @@ public:
         return (flags & FROM_SUBMITTER) == 0;
     }
 
+    void set_unknown_job_client_id( uint32_t clientId );
+    uint32_t unknown_job_client_id() const;
+    void set_job_id( uint32_t jobId );
+
     virtual void fill_from_channel(MsgChannel *c);
     virtual void send_to_channel(MsgChannel *c) const;
 
@@ -609,6 +663,7 @@ public:
     uint32_t out_uncompressed;
 
     uint32_t job_id;
+    uint32_t client_count; // number of CS -> C connections at the moment
 };
 
 class JobLocalBeginMsg : public Msg
@@ -644,7 +699,7 @@ public:
 class LoginMsg : public Msg
 {
 public:
-    LoginMsg(unsigned int myport, const std::string &_nodename, const std::string _host_platform);
+    LoginMsg(unsigned int myport, const std::string &_nodename, const std::string &_host_platform);
     LoginMsg()
         : Msg(M_LOGIN)
         , port(0) {}
@@ -681,8 +736,9 @@ class StatsMsg : public Msg
 public:
     StatsMsg()
         : Msg(M_STATS)
+        , load(0)
+        , client_count(0)
     {
-        load = 0;
     }
 
     virtual void fill_from_channel(MsgChannel *c);
@@ -703,6 +759,8 @@ public:
     uint32_t loadAvg5;
     uint32_t loadAvg10;
     uint32_t freeMem;
+
+    uint32_t client_count; // number of CS -> C connections at the moment
 };
 
 class EnvTransferMsg : public Msg
@@ -751,7 +809,7 @@ public:
     }
 
     MonGetCSMsg(int jobid, int hostid, GetCSMsg *m)
-        : GetCSMsg(Environments(), m->filename, m->lang, 1, m->target, 0, std::string(), false)
+        : GetCSMsg(Environments(), m->filename, m->lang, 1, m->target, 0, std::string(), false, m->client_count)
         , job_id(jobid)
         , clientid(hostid)
     {
